
from keras.models import Model
from keras.layers.merge import Concatenate, Multiply
from keras.layers import Activation, Input, Lambda
from keras.regularizers import l2
from keras.initializers import random_normal, constant
from keras import layers
from keras.layers import Dense, Flatten, Conv2D, UpSampling2D, Add, Conv2DTranspose
from keras.layers import MaxPooling2D
from keras.layers import GlobalMaxPooling2D
from keras.layers import ZeroPadding2D
from keras.layers import AveragePooling2D
from keras.layers import GlobalAveragePooling2D
from keras.layers import BatchNormalization
from keras.preprocessing import image
import keras.backend as K
from keras.utils import layer_utils
from keras.utils.data_utils import get_file
from keras.applications.imagenet_utils import decode_predictions
from keras.applications.imagenet_utils import preprocess_input
# from keras.applications.imagenet_utils import _obtain_input_shape
from keras.engine.topology import get_source_inputs
from keras.engine import Layer, InputSpec
from keras import initializers
from keras.layers import add

from . import mylayers
import code
import keras.backend as K

stages = 1
np_branch1 = 38
np_branch2 = 19
np_branch3 = 15


class Scale(Layer):
    """Custom Layer for ResNet used for BatchNormalization.
    
    Learns a set of weights and biases used for scaling the input data.
    the output consists simply in an element-wise multiplication of the input
    and a sum of a set of constants:

        out = in * gamma + beta,

    where 'gamma' and 'beta' are the weights and biases larned.

    Keyword arguments:
    axis -- integer, axis along which to normalize in mode 0. For instance,
        if your input tensor has shape (samples, channels, rows, cols),
        set axis to 1 to normalize per feature map (channels axis).
    momentum -- momentum in the computation of the exponential average 
        of the mean and standard deviation of the data, for 
        feature-wise normalization.
    weights -- Initialization weights.
        List of 2 Numpy arrays, with shapes:
        `[(input_shape,), (input_shape,)]`
    beta_init -- name of initialization function for shift parameter 
        (see [initializers](../initializers.md)), or alternatively,
        Theano/TensorFlow function to use for weights initialization.
        This parameter is only relevant if you don't pass a `weights` argument.
    gamma_init -- name of initialization function for scale parameter (see
        [initializers](../initializers.md)), or alternatively,
        Theano/TensorFlow function to use for weights initialization.
        This parameter is only relevant if you don't pass a `weights` argument.
        
    """
    def __init__(self, weights=None, axis=-1, momentum = 0.9, beta_init='zero', gamma_init='one', **kwargs):
        self.momentum = momentum
        self.axis = axis
        self.beta_init = initializers.get(beta_init)
        self.gamma_init = initializers.get(gamma_init)
        self.initial_weights = weights
        super(Scale, self).__init__(**kwargs)

    def build(self, input_shape):
        self.input_spec = [InputSpec(shape=input_shape)]
        shape = (int(input_shape[self.axis]),)

        self.gamma = K.variable(self.gamma_init(shape), name='%s_gamma'%self.name)
        self.beta = K.variable(self.beta_init(shape), name='%s_beta'%self.name)
        self.trainable = [self.gamma, self.beta]

        if self.initial_weights is not None:
            self.set_weights(self.initial_weights)
            del self.initial_weights

    def call(self, x, mask=None):
        input_shape = self.input_spec[0].shape
        broadcast_shape = [1] * len(input_shape)
        broadcast_shape[self.axis] = input_shape[self.axis]

        out = K.reshape(self.gamma, broadcast_shape) * x + K.reshape(self.beta, broadcast_shape)
        return out

    def get_config(self):
        config = {"momentum": self.momentum, "axis": self.axis}
        base_config = super(Scale, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))


def mytransform(source, ref_tensor):
    target_shape = K.shape(ref_tensor)
    source_shape = K.shape(source)
    return K.resize_images(source, target_shape[1]/source_shape[1],target_shape[2]/source_shape[2], "channels_last")
    # return tensorflow.image.resize_images(source, (target_shape[1],target_shape[2]))

def relu(x): return Activation('relu')(x)
def sigmoid(x): return Activation('sigmoid')(x)

def conv(x, nf, ks, name, weight_decay):
    # kernel_reg = l2(weight_decay[0]) if weight_decay else None
    # bias_reg = l2(weight_decay[1]) if weight_decay else None

    x = Conv2D(nf, (ks, ks), padding='same', name=name,
               kernel_regularizer=None,
               bias_regularizer=None,
               kernel_initializer=random_normal(stddev=0.01),
               bias_initializer=None)(x)
    return x

def conv_stride(x, nf, ks, name, weight_decay, stride=(2,2)):
    # kernel_reg = l2(weight_decay[0]) if weight_decay else None
    # bias_reg = l2(weight_decay[1]) if weight_decay else None

    x = Conv2D(nf, (ks, ks), padding='same', name=name, strides=stride,
               kernel_regularizer=None,
               bias_regularizer=None,
               kernel_initializer=random_normal(stddev=0.01),
               bias_initializer=None)(x)
    return x

def pooling(x, ks, st, name):
    x = MaxPooling2D((ks, ks), strides=(st, st), name=name)(x)
    return x

def identity_block(input_tensor, kernel_size, filters, stage, block):
    """The identity_block is the block that has no conv layer at shortcut
    
    Keyword arguments
    input_tensor -- input tensor
    kernel_size -- defualt 3, the kernel size of middle conv layer at main path
    filters -- list of integers, the nb_filters of 3 conv layer at main path
    stage -- integer, current stage label, used for generating layer names
    block -- 'a','b'..., current block label, used for generating layer names
    
    """
    eps = 1.1e-5
    
    # if K.image_dim_ordering() == 'tf':
    bn_axis = 3
    # else:
        # bn_axis = 1
    
    nb_filter1, nb_filter2, nb_filter3 = filters
    conv_name_base = 'res' + str(stage) + block + '_branch'
    bn_name_base = 'bn' + str(stage) + block + '_branch'
    scale_name_base = 'scale' + str(stage) + block + '_branch'

    x = Conv2D(nb_filter1, (1, 1), name=conv_name_base + '2a',use_bias=False)(input_tensor)
    x = BatchNormalization(epsilon=eps, axis=bn_axis,name=bn_name_base + '2a')(x)
    x = Scale(axis=bn_axis, name=scale_name_base + '2a')(x)
    x = Activation('relu', name=conv_name_base + '2a_relu')(x)

    x = ZeroPadding2D((1, 1), name=conv_name_base + '2b_zeropadding')(x)
    x = Conv2D(nb_filter2, (kernel_size, kernel_size),name=conv_name_base + '2b', use_bias=False)(x)
    x = BatchNormalization(epsilon=eps, axis=bn_axis,name=bn_name_base + '2b')(x)
    x = Scale(axis=bn_axis, name=scale_name_base + '2b')(x)
    x = Activation('relu', name=conv_name_base + '2b_relu')(x)

    x = Conv2D(nb_filter3, (1, 1), name=conv_name_base + '2c',use_bias=False)(x)
    x = BatchNormalization(epsilon=eps, axis=bn_axis,name=bn_name_base + '2c')(x)
    x = Scale(axis=bn_axis, name=scale_name_base + '2c')(x)

    x = add([x, input_tensor], name='res' + str(stage) + block)
    x = Activation('relu', name='res' + str(stage) + block + '_relu')(x)
    return x


def conv_block(input_tensor, kernel_size, filters, stage, block, strides=(2, 2)):
    """conv_block is the block that has a conv layer at shortcut
    
    Keyword arguments:
    input_tensor -- input tensor
    kernel_size -- defualt 3, the kernel size of middle conv layer at main path
    filters -- list of integers, the nb_filters of 3 conv layer at main path
    stage -- integer, current stage label, used for generating layer names
    block -- 'a','b'..., current block label, used for generating layer names
        
    Note that from stage 3, the first conv layer at main path is with subsample=(2,2)
    And the shortcut should have subsample=(2,2) as well
    
    """
    eps = 1.1e-5
    
    # if K.image_dim_ordering() == 'tf':
    bn_axis = 3
    # else:
        # bn_axis = 1
    
    nb_filter1, nb_filter2, nb_filter3 = filters
    conv_name_base = 'res' + str(stage) + block + '_branch'
    bn_name_base = 'bn' + str(stage) + block + '_branch'
    scale_name_base = 'scale' + str(stage) + block + '_branch'

    x = Conv2D(nb_filter1, (1, 1), strides=strides,name=conv_name_base + '2a', use_bias=False)(input_tensor)
    x = BatchNormalization(epsilon=eps, axis=bn_axis,name=bn_name_base + '2a')(x)
    x = Scale(axis=bn_axis, name=scale_name_base + '2a')(x)
    x = Activation('relu', name=conv_name_base + '2a_relu')(x)

    x = ZeroPadding2D((1, 1), name=conv_name_base + '2b_zeropadding')(x)
    x = Conv2D(nb_filter2, (kernel_size, kernel_size),name=conv_name_base + '2b', use_bias=False)(x)
    x = BatchNormalization(epsilon=eps, axis=bn_axis,name=bn_name_base + '2b')(x)
    x = Scale(axis=bn_axis, name=scale_name_base + '2b')(x)
    x = Activation('relu', name=conv_name_base + '2b_relu')(x)

    x = Conv2D(nb_filter3, (1, 1),name=conv_name_base + '2c', use_bias=False)(x)
    x = BatchNormalization(epsilon=eps, axis=bn_axis,name=bn_name_base + '2c')(x)
    x = Scale(axis=bn_axis, name=scale_name_base + '2c')(x)

    shortcut = Conv2D(nb_filter3, (1, 1), strides=strides,name=conv_name_base + '1', use_bias=False)(input_tensor)
    shortcut = BatchNormalization(epsilon=eps, axis=bn_axis,name=bn_name_base + '1')(shortcut)
    shortcut = Scale(axis=bn_axis, name=scale_name_base + '1')(shortcut)

    x = add([x, shortcut], name='res' + str(stage) + block)
    x = Activation('relu', name='res' + str(stage) + block + '_relu')(x)
    return x


def ResNet101_graph(img_input, weight_decay):
    eps = 1.1e-5

    branch = 0
    # if K.image_data_format() == 'channels_last':
    bn_axis = 3
    # else:
        # bn_axis = 1
    # C1 --------------------------------------------------
    x = ZeroPadding2D((3, 3))(img_input)
    x = Conv2D(64, (7, 7), strides=(2, 2), name='conv1', use_bias=False)(x)
    x = BatchNormalization(epsilon=eps, axis=bn_axis, name='bn_conv1')(x)
    x = Scale(axis=bn_axis, name='scale_conv1')(x)
    x = Activation('relu', name='conv1_relu')(x)
    x = MaxPooling2D((3, 3), strides=(2, 2), name='pool1')(x)

    C1 = x

    # C2 --------------------------------------------------
    x = conv_block(x, 3, [64, 64, 256], stage=2, block='a', strides=(1, 1))
    x = identity_block(x, 3, [64, 64, 256], stage=2, block='b')
    x = identity_block(x, 3, [64, 64, 256], stage=2, block='c')

    C2 = x

    # C3 --------------------------------------------------
    x = conv_block(x, 3, [128, 128, 512], stage=3, block='a')
    for i in range(1, 3):
        x = identity_block(x, 3, [128, 128, 512], stage=3, block='b' + str(i))
    C3 = x

    # C4 --------------------------------------------------
    x = conv_block(x, 3, [256, 256, 1024], stage=4, block='a')
    for i in range(1, 23):
        x = identity_block(x, 3, [256, 256, 1024], stage=4, block='b' + str(i))
    C4 = x

    # C5 ---------------------------------------------------
    x = conv_block(x, 3, [512, 512, 2048], stage=5, block='a')
    x = identity_block(x, 3, [512, 512, 2048], stage=5, block='b')
    x = identity_block(x, 3, [512, 512, 2048], stage=5, block='c')
    C5 = x

    return C1, C2, C3, C4, C5


def create_pyramid_features(C1, C2, C3, C4, C5, feature_size=256):
    P5 = Conv2D(feature_size, kernel_size=1, strides=1, padding='same', name='C5_reduced')(C5)
    P4 = Conv2D(feature_size, kernel_size=1, strides=1, padding='same', name='C4_reduced')(C4)
    P3 = Conv2D(feature_size, kernel_size=1, strides=1, padding='same', name='C3_reduced')(C3)
    P2 = Conv2D(feature_size, kernel_size=1, strides=1, padding='same', name='C2_reduced')(C2)
    P1 = Conv2D(feature_size, kernel_size=1, strides=1, padding='same', name='C1_reduced')(C1)

    # upsample P5 to get P5_up1
    P5_up1 = mylayers.DeformableDeConv(name='P5_up1_deconv',
                                             kernel_size=4,
                                             stride=[1,2,2,1],
                                             filter_num=feature_size)([P5,P4])
    # upsample P5_up1 to get P5_up2
    P5_up2 = mylayers.DeformableDeConv(name='P5_up2_deconv',
                                             kernel_size=4,
                                             stride=[1,2,2,1],
                                             filter_num=feature_size)([P5_up1,P3])
    # upsample P4 to get P4_up1
    P4_up1 = mylayers.DeformableDeConv(name='P4_up1_deconv',
                                             kernel_size=4,
                                             stride=[1,2,2,1],
                                             filter_num=feature_size)([P4,P3])

    # downsample P1 to get P1_down1
    P1_down1 = Conv2D(feature_size, kernel_size=1, strides=1, padding='same', name='P1_down1')(P1)
    # downsample P1_down1 to get P1_down2
    P1_down2 = Conv2D(feature_size, kernel_size=1, strides=2, padding='same', name='P1_down2')(P1_down1)
    # downsample P2 to get P2_down
    P2_down1 = Conv2D(feature_size, kernel_size=1, strides=2, padding='same', name='P2_down1')(P2)


    P5_up2 = Conv2D(feature_size, kernel_size=3, strides=1, padding='same', name='P5_up2_head')(P5_up2)
    P5_up2 = relu(P5_up2)

    P4_up1 = Conv2D(feature_size, kernel_size=3, strides=1, padding='same', name='P4_up1_head')(P4_up1)
    P4_up1 = relu(P4_up1)

    P3 = Conv2D(feature_size, kernel_size=3, strides=1, padding='same', name='P3_head')(P3)
    P3 = relu(P3)

    P2_down1 = Conv2D(feature_size, kernel_size=3, strides=1, padding='same', name='P2_down1_head')(P2_down1)
    P2_down1 = relu(P2_down1)

    P1_down2 = Conv2D(feature_size, kernel_size=3, strides=1, padding='same', name='P1_down2_head')(P1_down2)
    P1_down2 = relu(P1_down2)


    # Concatenate features at different levels
    pyramid_feat = []
    pyramid_feat.append(P5_up2)
    pyramid_feat.append(P4_up1)
    pyramid_feat.append(P3)
    pyramid_feat.append(P2_down1)
    pyramid_feat.append(P1_down2)
    feats = Concatenate()(pyramid_feat)

    return feats



def stage1_block(x, num_p, branch, weight_decay):
    # Block 1
    x = conv(x, 512, 3, "Mconv1_stage1_L%d" % branch, (weight_decay, 0))
    x = relu(x)
    x = conv(x, 512, 3, "Mconv2_stage1_L%d" % branch, (weight_decay, 0))
    x = relu(x)
    x = conv(x, 512, 3, "Mconv3_stage1_L%d" % branch, (weight_decay, 0))
    x = relu(x)
    x = conv(x, 512, 3, "Mconv4_stage1_L%d" % branch, (weight_decay, 0))
    x = relu(x)
    x = conv(x, 512, 3, "Mconv5_stage1_L%d" % branch, (weight_decay, 0))
    x = relu(x)
    x = conv(x, 512, 3, "Mconv6_stage1_L%d" % branch, (weight_decay, 0))
    x = relu(x)
    x = conv(x, 512, 1, "Mconv7_stage1_L%d" % branch, (weight_decay, 0))
    x = relu(x)
    x = conv(x, num_p, 1, "Mconv8_stage1_L%d" % branch, (weight_decay, 0))

    return x



def stage1_segmentation_block(x, num_p, branch, weight_decay):
    # Block 1
    x = conv(x, 256, 3, "Mconv1_stage1_L%d" % branch, (weight_decay, 0))
    x = relu(x)
    x = conv(x, 256, 3, "Mconv2_stage1_L%d" % branch, (weight_decay, 0))
    x = relu(x)
    x = conv(x, 256, 3, "Mconv3_stage1_L%d" % branch, (weight_decay, 0))
    x = relu(x)
    x = conv(x, 256, 3, "Mconv4_stage1_L%d" % branch, (weight_decay, 0))
    x = relu(x)
    x = conv(x, 256, 1, "Mconv5_stage1_L%d" % branch, (weight_decay, 0))
    x = relu(x)
    x = conv(x, num_p, 1, "Mconv6_stage1_L%d" % branch, (weight_decay, 0))
    #x = sigmoid(x)
    x = Activation('softmax')(x)

    return x

def apply_mask(x, mask1, mask2, mask3, num_p, stage, branch):
    w_name = "weight_stage%d_L%d" % (stage, branch)
    if num_p == np_branch1:
        w = Multiply(name=w_name)([x, mask1])  # vec_weight
    elif num_p == np_branch2:
        w = Multiply(name=w_name)([x, mask2])  # vec_heat
    elif num_p == np_branch3:
        w = Multiply(name=w_name)([x, mask3])  # seg
    else:
        assert False, "wrong number of layers num_p=%d " % num_p
    return w


def get_training_model_resnet101(weight_decay, gpus=None):

    img_input_shape = (None, None, 3)
    vec_input_shape = (None, None, 38)
    heat_input_shape = (None, None, 19)
    seg_input_shape = (None, None, 15)

    inputs = []
    outputs = []

    img_input = Input(shape=img_input_shape)
    vec_weight_input = Input(shape=vec_input_shape)
    heat_weight_input = Input(shape=heat_input_shape)
    seg_weight_input = Input(shape=seg_input_shape)

    inputs.append(img_input)
    inputs.append(vec_weight_input)
    inputs.append(heat_weight_input)
    inputs.append(seg_weight_input)

    # resnet101
    C1, C2, C3, C4, C5 = ResNet101_graph(img_input, weight_decay)

    stage0_out = create_pyramid_features(C1, C2, C3, C4, C5)

    # Additional layers for learning multi-scale semantics
    stage0_out = conv(stage0_out, 512, 3, "pyramid_1_CPM", (weight_decay, 0))
    stage0_out = relu(stage0_out)
    stage0_out = conv(stage0_out, 512, 3, "pyramid_2_CPM", (weight_decay, 0))
    stage0_out = relu(stage0_out)

    # stage 1 - branch 1 (PAF)
    stage1_branch1_out = stage1_block(stage0_out, np_branch1, 1, weight_decay)
    w1 = apply_mask(stage1_branch1_out, vec_weight_input, heat_weight_input, seg_weight_input, np_branch1, 1, 1)

    # stage 1 - branch 2 (confidence maps)
    stage1_branch2_out = stage1_block(stage0_out, np_branch2, 2, weight_decay)
    w2 = apply_mask(stage1_branch2_out, vec_weight_input, heat_weight_input, seg_weight_input, np_branch2, 1, 2)

    # stage 1 - branch 3 (semantic segmentation)
    stage1_branch3_out = stage1_segmentation_block(stage0_out, np_branch3, 3, weight_decay)
    w3 = apply_mask(stage1_branch3_out, vec_weight_input, heat_weight_input, seg_weight_input, np_branch3, 1, 3)

    outputs.append(w1)
    outputs.append(w2)
    outputs.append(w3)


    if gpus is None:
        model = Model(inputs=inputs, outputs=outputs)
    else:
        import tensorflow as tf
        with tf.device('/cpu:0'): #this model will not be actually used, it's template
            model = Model(inputs=inputs, outputs=outputs)

    return model



def get_testing_model_resnet101():

    img_input_shape = (None, None, 3)

    img_input = Input(shape=img_input_shape)

    C1, C2, C3, C4, C5 = ResNet101_graph(img_input, None)

    stage0_out = create_pyramid_features(C1, C2, C3, C4, C5)

    # Additional layers for learning multi-scale semantics
    stage0_out = conv(stage0_out, 512, 3, "pyramid_1_CPM", (None, 0))
    stage0_out = relu(stage0_out)
    stage0_out = conv(stage0_out, 512, 3, "pyramid_2_CPM", (None, 0))
    stage0_out = relu(stage0_out)

    # stage 1 - branch 1 (PAF)
    stage1_branch1_out = stage1_block(stage0_out, np_branch1, 1, None)

    # stage 1 - branch 2 (confidence maps)
    stage1_branch2_out = stage1_block(stage0_out, np_branch2, 2, None)

    # stage 1 - branch 3 (semantic segmentation)
    stage1_branch3_out = stage1_segmentation_block(stage0_out, np_branch3, 3, None)


    model = Model(inputs=[img_input], outputs=[stage1_branch1_out, stage1_branch2_out, stage1_branch3_out])
    return model




"""
MIT License

Copyright (c) 2020 Adeel <kingadeel2017@outlook.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from xtreme_vision.Detection.centernet import PoseEstimation as estimate
from xtreme_vision.Segmentation.cdcl.inference_15parts_skeletons import run_image, run_video
import numpy as np
import cv2
import sys
from PIL import Image


class Pose_Estimation:
    
    """
    This is Pose-Estimation Class in Xtreme-Vision Library, it provides the support of State-Of-The-Art Models 
    like CenterNet. After Instantiating this Class, you can set its properties and use pre-defined
    functions for performing segmentation Tasks out of the box.

        Use_CenterNet()                             # To Specify which Model to Use
        Detect_From_Image()                         # To Detect from Images
        Detect_From_Video()                         # To Detect from Videos

    """

    def __init__(self):
    
        self.model = None
        self.modelLoaded = False
        self.modelType = None
  
    def Use_CenterNet(self, weights_path:str = None):
        
        """[This function is used to set the Model Type to CenterNet, Automatically downloads the 
        weights file and Loads the Model.]
        """

        self.model = estimate()
        print('-' * 20)
        print('Loading the Model \n Please Wait...')
        print('-' * 20)
        self.model.load_model(weights_path)
        self.modelLoaded = True
        self.modelType = 'centernet'   

    
    def Use_CDCL(self):

        self.modelLoaded = True
        self.modelType = 'cdcl'
    
    def Detect_From_Image(self, input_path:str, output_path:str):
        
        """[This Function is used to detect pose from Images.]
        
        Args:
            input_path: (str) [path to the input image with jpg/jpeg/png extension]
            output_path: (str) [path to save the output image with jpg/jpeg/png extension]
 
        Raises:
            RuntimeError: [If Model is not Loaded before using this Function]
            RuntimeError: [If any othre Model type is specified other than CenterNet]
        """
        
        if self.modelLoaded != True:
            raise RuntimeError ('Before calling this function, you have to call Use_CenterNet().')
              
        img = np.array(Image.open(input_path))[..., ::-1]
              
        if self.modelType == 'centernet':
            
            _ = self.model.predict(img = img, output_path = output_path, debug = True) 
                  
        elif self.modelType == 'cdcl':

            _ = run_image(input_path, output_path)
        
        else:
            raise RuntimeError ('Invalid ModelType: Valid Type Is "CenterNet"')

    def Detect_From_Video(self, input_path:str, output_path:str, fps:int = 25):
        
        """[This Function is used to detect pose from Videos.]
        
        Args:
            input_path: (str) [path to the input Video with mp4/avi extension]
            output_path: (str) [path to save the output Video with mp4/avi extension]
            fps: (int) [frames per second for video processing]
 
        Raises:
            RuntimeError: [If Model is not Loaded before using this Function]
            RuntimeError: [If any othre Model type is specified other than CenterNet]
        """
    
        if self.modelLoaded != True:
            raise RuntimeError ('Before calling this function, you have to call Use_CenterNet().')

        if self.modelType == 'cdcl':
            _ = run_video(input_path, output_path, fps)
            sys.exit()
        
        out = None
        cap = cv2.VideoCapture(input_path)
        length = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
        print(f'\nThere are {length} Frames in this video')
        print('-' * 20)
        print('Detecting Objects in the Video... Please Wait...')
        print('-' * 20)

        while(cap.isOpened()):
            retreive, frame = cap.read()
            if not retreive:
                break

            frame = np.array(frame)[..., ::-1]      
            if self.modelType == 'centernet':
                im = self.model.predict(img = frame, output_path='./', debug = False)
                im = cv2.cvtColor(np.array(im), cv2.COLOR_RGB2BGR)

            else:
                raise RuntimeError ('Invalid ModelType: Valid Type Is "CenterNet"')
            
            if out is None:
                fourcc = cv2.VideoWriter_fourcc(*'DIVX')
                out = cv2.VideoWriter(output_path, fourcc, fps, (frame.shape[1], frame.shape[0]))
            
            out.write(im)
        print('Done. Processing has been Finished... Please Check Output Video.')
        out.release()
        cap.release()



# -*- coding: utf-8 -*-

"""
keras_resnet.classifiers
~~~~~~~~~~~~~~~~~~~~~~~~

This module implements popular residual classifiers.
"""

from xtreme_vision.Detection.keras_resnet.classifiers._2d import (
    ResNet18,
    ResNet34,
    ResNet50,
    ResNet101,
    ResNet152,
    ResNet200
)



from configobj import ConfigObj
import numpy as np


def config_reader():
    config = ConfigObj('xtreme_vision/Segmentation/cdcl/config')

    param = config['param']
    model_id = param['modelID']
    model = config['models'][model_id]
    model['boxsize'] = int(model['boxsize'])
    model['stride'] = int(model['stride'])
    model['padValue'] = int(model['padValue'])
    #param['starting_range'] = float(param['starting_range'])
    #param['ending_range'] = float(param['ending_range'])
    param['octave'] = int(param['octave'])
    param['use_gpu'] = int(param['use_gpu'])
    param['starting_range'] = float(param['starting_range'])
    param['ending_range'] = float(param['ending_range'])
    param['scale_search'] = map(float, param['scale_search'])
    param['thre1'] = float(param['thre1'])
    param['thre2'] = float(param['thre2'])
    param['thre3'] = float(param['thre3'])
    param['mid_num'] = int(param['mid_num'])
    param['min_num'] = int(param['min_num'])
    param['crop_ratio'] = float(param['crop_ratio'])
    param['bbox_ratio'] = float(param['bbox_ratio'])
    param['GPUdeviceNumber'] = int(param['GPUdeviceNumber'])

    return param, model

if __name__ == "__main__":
    config_reader()



# -*- coding: utf-8 -*-

"""
keras_resnet.models
~~~~~~~~~~~~~~~~~~~

This module implements popular residual models.
"""

from xtreme_vision.Detection.keras_resnet.models._1d import (
    ResNet1D,
    ResNet1D18,
    ResNet1D34,
    ResNet1D50,
    ResNet1D101,
    ResNet1D152,
    ResNet1D200
)

from xtreme_vision.Detection.keras_resnet.models._2d import (
    ResNet2D,
    ResNet2D18,
    ResNet2D34,
    ResNet2D50,
    ResNet2D101,
    ResNet2D152,
    ResNet2D200
)

from xtreme_vision.Detection.keras_resnet.models._3d import (
    ResNet3D,
    ResNet3D18,
    ResNet3D34,
    ResNet3D50,
    ResNet3D101,
    ResNet3D152,
    ResNet3D200
)

from xtreme_vision.Detection.keras_resnet.models._feature_pyramid_2d import (
    FPN2D,
    FPN2D18,
    FPN2D34,
    FPN2D50,
    FPN2D101,
    FPN2D152,
    FPN2D200
)

from xtreme_vision.Detection.keras_resnet.models._time_distributed_2d import (
    TimeDistributedResNet,
    TimeDistributedResNet18,
    TimeDistributedResNet34,
    TimeDistributedResNet50,
    TimeDistributedResNet101,
    TimeDistributedResNet152,
    TimeDistributedResNet200
)

# for backwards compatibility reasons
ResNet = ResNet2D
ResNet18 = ResNet2D18
ResNet34 = ResNet2D34
ResNet50 = ResNet2D50
ResNet101 = ResNet2D101
ResNet152 = ResNet2D152
ResNet200 = ResNet2D200



import numpy as np
from io import StringIO
import PIL.Image
from IPython.display import Image, display

def showBGRimage(a, fmt='jpeg'):
    a = np.uint8(np.clip(a, 0, 255))
    a[:,:,[0,2]] = a[:,:,[2,0]] # for B,G,R order
    f = StringIO()
    PIL.Image.fromarray(a).save(f, fmt)
    display(Image(data=f.getvalue()))

def showmap(a, fmt='png'):
    a = np.uint8(np.clip(a, 0, 255))
    f = StringIO()
    PIL.Image.fromarray(a).save(f, fmt)
    display(Image(data=f.getvalue()))

#def checkparam(param):
#    octave = param['octave']
#    starting_range = param['starting_range']
#    ending_range = param['ending_range']
#    assert starting_range <= ending_range, 'starting ratio should <= ending ratio'
#    assert octave >= 1, 'octave should >= 1'
#    return starting_range, ending_range, octave

def getJetColor(v, vmin, vmax):
    c = np.zeros((3))
    if (v < vmin):
        v = vmin
    if (v > vmax):
        v = vmax
    dv = vmax - vmin
    if (v < (vmin + 0.125 * dv)): 
        c[0] = 256 * (0.5 + (v * 4)) #B: 0.5 ~ 1
    elif (v < (vmin + 0.375 * dv)):
        c[0] = 255
        c[1] = 256 * (v - 0.125) * 4 #G: 0 ~ 1
    elif (v < (vmin + 0.625 * dv)):
        c[0] = 256 * (-4 * v + 2.5)  #B: 1 ~ 0
        c[1] = 255
        c[2] = 256 * (4 * (v - 0.375)) #R: 0 ~ 1
    elif (v < (vmin + 0.875 * dv)):
        c[1] = 256 * (-4 * v + 3.5)  #G: 1 ~ 0
        c[2] = 255
    else:
        c[2] = 256 * (-4 * v + 4.5) #R: 1 ~ 0.5                      
    return c

def colorize(gray_img):
    out = np.zeros(gray_img.shape + (3,))
    for y in range(out.shape[0]):
        for x in range(out.shape[1]):
            out[y,x,:] = getJetColor(gray_img[y,x], 0, 1)
    return out

def padRightDownCorner(img, stride, padValue):
    h = img.shape[0]
    w = img.shape[1]

    pad = 4 * [None]
    pad[0] = 0 # up
    pad[1] = 0 # left
    pad[2] = 0 if (h%stride==0) else stride - (h % stride) # down
    pad[3] = 0 if (w%stride==0) else stride - (w % stride) # right

    img_padded = img
    pad_up = np.tile(img_padded[0:1,:,:]*0 + padValue, (pad[0], 1, 1))
    img_padded = np.concatenate((pad_up, img_padded), axis=0)
    pad_left = np.tile(img_padded[:,0:1,:]*0 + padValue, (1, pad[1], 1))
    img_padded = np.concatenate((pad_left, img_padded), axis=1)
    pad_down = np.tile(img_padded[-2:-1,:,:]*0 + padValue, (pad[2], 1, 1))
    img_padded = np.concatenate((img_padded, pad_down), axis=0)
    pad_right = np.tile(img_padded[:,-2:-1,:]*0 + padValue, (1, pad[3], 1))
    img_padded = np.concatenate((img_padded, pad_right), axis=1)

    return img_padded, pad



"""
MIT License

Copyright (c) 2020 Adeel <kingadeel2017@outlook.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


from xtreme_vision.Detection.yolov4.tf import YOLOv4 as yolo_main
import numpy as np
import cv2

labels = {0: 'person', 1: 'bicycle', 2: 'car', 3: 'motorcycle', 4: 'airplane', 5: 'bus', 6: 'train', 7: 'truck', 8: 'boat',
            9: 'traffic light', 10: 'fire hydrant', 11: 'stop sign', 12: 'parking meter', 13: 'bench', 14: 'bird', 15: 'cat', 16: 'dog',
            17: 'horse', 18: 'sheep', 19: 'cow', 20: 'elephant', 21: 'bear', 22: 'zebra', 23: 'giraffe', 24: 'backpack', 25: 'umbrella',
            26: 'handbag', 27: 'tie', 28: 'suitcase', 29: 'frisbee', 30: 'skis', 31: 'snowboard', 32: 'sports ball', 33: 'kite',
            34: 'baseball bat', 35: 'baseball glove', 36: 'skateboard', 37: 'surfboard', 38: 'tennis racket', 39: 'bottle', 40: 'wine glass',
            41: 'cup', 42: 'fork', 43: 'knife', 44: 'spoon', 45: 'bowl', 46: 'banana', 47: 'apple', 48: 'sandwich', 49: 'orange',
            50: 'broccoli', 51: 'carrot', 52: 'hot dog', 53: 'pizza', 54: 'donut', 55: 'cake', 56: 'chair', 57: 'couch', 58: 'potted plant',
            59: 'bed', 60: 'dining table', 61: 'toilet', 62: 'tv', 63: 'laptop', 64: 'mouse', 65: 'remote', 66: 'keyboard', 67: 'cell phone',
            68: 'microwave', 69: 'oven', 70: 'toaster', 71: 'sink', 72: 'refrigerator', 73: 'book', 74: 'clock', 75: 'vase', 76: 'scissors',
            77: 'teddy bear', 78: 'hair drier', 79: 'toothbrush'}

class YOLOv4:
    def __init__(self):
        self.weights_path = ""
        self.model = None
        self.yolo_classes = ""
        self.iou = 0
        self.score = 0
        self.input_shape = 0
        self.output_path = ""
    
    def load_model(self, weights_path:str = None, classes_path:str = None, input_shape:int = 608):
        if (weights_path  is None) or (classes_path is None):
            raise RuntimeError ('weights_path AND classes_path should not be None.')
        
        self.yolo_classes = classes_path
        self.weights_path = weights_path
        self.input_shape = input_shape
        
        self.model = yolo_main(shape = self.input_shape)
        self.model.classes = self.yolo_classes
        self.model.make_model()
        self.model.load_weights(self.weights_path, weights_type = 'yolo')
        
    def predict(self, img:np.ndarray, output_path:str, iou = 0.45, score = 0.25, custom_objects:dict = None, 
                debug=True):

        self.output_path = output_path
        self.iou = iou
        self.score = score  
        #img = np.array(Image.open(img))[..., ::-1] 
        pred_bboxes = self.model.predict(img, iou_threshold = self.iou, score_threshold = self.score)
        boxes = []
        if (custom_objects != None):
          for i in range(len(pred_bboxes)):
            check_name = labels[pred_bboxes[i][4]]
            check = custom_objects.get(check_name, 'invalid')
            if check == 'invalid':
              continue
            elif check == 'valid':
              boxes.append(list(pred_bboxes[i]))
          boxes = np.array(boxes)
          res = self.model.draw_bboxes(img, boxes)
          if debug:
            cv2.imwrite(self.output_path, res)

        else:
          res = self.model.draw_bboxes(img, pred_bboxes)
          if debug:
              cv2.imwrite(self.output_path, res)
        
        return res
  
class TinyYOLOv4:
    def __init__(self):
        self.weights_path = ""
        self.model = None
        self.yolo_classes = ""
        self.iou = 0
        self.score = 0
        self.input_shape = 0
        self.output_path = ""
    
    def load_model(self, weights_path:str = None, classes_path:str = None, input_shape:int = 0):
        if (weights_path  is None) or (classes_path is None):
            raise RuntimeError ('weights_path AND classes_path should not be None.')
        
        self.yolo_classes = classes_path
        self.weights_path = weights_path
        self.input_shape = input_shape
        self.model = yolo_main(tiny = True, shape = self.input_shape)
        self.model.classes = self.yolo_classes
        self.model.make_model()
        self.model.load_weights(self.weights_path, weights_type = 'yolo')
        
    def predict(self, img:np.ndarray, output_path:str, iou = 0.4, score = 0.07, custom_objects:dict = None,
                debug=True):

        self.output_path = output_path
        self.iou = iou
        self.score = score  
        #img = np.array(Image.open(img))[..., ::-1] 
        pred_bboxes = self.model.predict(img, iou_threshold = self.iou, score_threshold = self.score)
        
        boxes = []
        if (custom_objects != None):
          for i in range(len(pred_bboxes)):
            check_name = labels[pred_bboxes[i][4]]
            check = custom_objects.get(check_name, 'invalid')
            if check == 'invalid':
              continue
            elif check == 'valid':
              boxes.append(list(pred_bboxes[i]))
          boxes = np.array(boxes)
          res = self.model.draw_bboxes(img, boxes)
          if debug:
            cv2.imwrite(self.output_path, res)

        else:
          res = self.model.draw_bboxes(img, pred_bboxes)
          if debug:
              cv2.imwrite(self.output_path, res)
        
        return res
        


"""
Copyright 2017-2018 lvaleriu (https://github.com/lvaleriu/)

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import csv
import os.path

import numpy as np
from PIL import Image

from .generator import Generator
from ..utils.image import read_image_bgr

kitti_classes = {
    'Car': 0,
    'Van': 1,
    'Truck': 2,
    'Pedestrian': 3,
    'Person_sitting': 4,
    'Cyclist': 5,
    'Tram': 6,
    'Misc': 7,
    'DontCare': 7
}


class KittiGenerator(Generator):
    """ Generate data for a KITTI dataset.

    See http://www.cvlibs.net/datasets/kitti/ for more information.
    """

    def __init__(
        self,
        base_dir,
        subset='train',
        **kwargs
    ):
        """ Initialize a KITTI data generator.

        Args
            base_dir: Directory w.r.t. where the files are to be searched (defaults to the directory containing the csv_data_file).
            subset: The subset to generate data for (defaults to 'train').
        """
        self.base_dir = base_dir

        label_dir = os.path.join(self.base_dir, subset, 'labels')
        image_dir = os.path.join(self.base_dir, subset, 'images')

        """
        1    type         Describes the type of object: 'Car', 'Van', 'Truck',
                             'Pedestrian', 'Person_sitting', 'Cyclist', 'Tram',
                             'Misc' or 'DontCare'
        1    truncated    Float from 0 (non-truncated) to 1 (truncated), where
                         truncated refers to the object leaving image boundaries
        1    occluded     Integer (0,1,2,3) indicating occlusion state:
                         0 = fully visible, 1 = partly occluded
                         2 = largely occluded, 3 = unknown
        1    alpha        Observation angle of object, ranging [-pi..pi]
        4    bbox         2D bounding box of object in the image (0-based index):
                         contains left, top, right, bottom pixel coordinates
        3    dimensions   3D object dimensions: height, width, length (in meters)
        3    location     3D object location x,y,z in camera coordinates (in meters)
        1    rotation_y   Rotation ry around Y-axis in camera coordinates [-pi..pi]
        """

        self.labels = {}
        self.classes = kitti_classes
        for name, label in self.classes.items():
            self.labels[label] = name

        self.image_data = dict()
        self.images = []
        for i, fn in enumerate(os.listdir(label_dir)):
            label_fp = os.path.join(label_dir, fn)
            image_fp = os.path.join(image_dir, fn.replace('.txt', '.png'))

            self.images.append(image_fp)

            fieldnames = ['type', 'truncated', 'occluded', 'alpha', 'left', 'top', 'right', 'bottom', 'dh', 'dw', 'dl',
                          'lx', 'ly', 'lz', 'ry']
            with open(label_fp, 'r') as csv_file:
                reader = csv.DictReader(csv_file, delimiter=' ', fieldnames=fieldnames)
                boxes = []
                for line, row in enumerate(reader):
                    label = row['type']
                    cls_id = kitti_classes[label]

                    annotation = {'cls_id': cls_id, 'x1': row['left'], 'x2': row['right'], 'y2': row['bottom'], 'y1': row['top']}
                    boxes.append(annotation)

                self.image_data[i] = boxes

        super(KittiGenerator, self).__init__(**kwargs)

    def size(self):
        """ Size of the dataset.
        """
        return len(self.images)

    def num_classes(self):
        """ Number of classes in the dataset.
        """
        return max(self.classes.values()) + 1

    def has_label(self, label):
        """ Return True if label is a known label.
        """
        return label in self.labels

    def has_name(self, name):
        """ Returns True if name is a known class.
        """
        return name in self.classes

    def name_to_label(self, name):
        """ Map name to label.
        """
        raise NotImplementedError()

    def label_to_name(self, label):
        """ Map label to name.
        """
        return self.labels[label]

    def image_aspect_ratio(self, image_index):
        """ Compute the aspect ratio for an image with image_index.
        """
        # PIL is fast for metadata
        image = Image.open(self.images[image_index])
        return float(image.width) / float(image.height)

    def image_path(self, image_index):
        """ Get the path to an image.
        """
        return self.images[image_index]

    def load_image(self, image_index):
        """ Load an image at the image_index.
        """
        return read_image_bgr(self.image_path(image_index))

    def load_annotations(self, image_index):
        """ Load annotations for an image_index.
        """
        image_data = self.image_data[image_index]
        annotations = {'labels': np.empty((len(image_data),)), 'bboxes': np.empty((len(image_data), 4))}

        for idx, ann in enumerate(image_data):
            annotations['bboxes'][idx, 0] = float(ann['x1'])
            annotations['bboxes'][idx, 1] = float(ann['y1'])
            annotations['bboxes'][idx, 2] = float(ann['x2'])
            annotations['bboxes'][idx, 3] = float(ann['y2'])
            annotations['labels'][idx] = int(ann['cls_id'])

        return annotations



import cv2
import math
import time
import numpy as np
from . import util
from .config_reader import config_reader
from scipy.ndimage.filters import gaussian_filter
from tensorflow.keras.models import load_model
import tensorflow as tf
import code
import copy
import scipy.ndimage as sn
from PIL import Image
from tqdm import tqdm
from .model_simulated_RGB101 import get_testing_model_resnet101
from .human_seg.human_seg_gt import human_seg_combine_argmax


right_part_idx = [2, 3, 4,  8,  9, 10, 14, 16]
left_part_idx =  [5, 6, 7, 11, 12, 13, 15, 17]
human_part = [0,1,2,4,3,6,5,8,7,10,9,12,11,14,13]
human_ori_part = [0,1,2,3,4,5,6,7,8,9,10,11,12,13,14]
seg_num = 15 # current model supports 15 parts only


# find connection in the specified sequence, center 29 is in the position 15
limbSeq = [[2, 3], [2, 6], [3, 4], [4, 5], [6, 7], [7, 8], [2, 9], [9, 10], \
           [10, 11], [2, 12], [12, 13], [13, 14], [2, 1], [1, 15], [15, 17], \
           [1, 16], [16, 18], [3, 17], [6, 18]]

# the middle joints heatmap correpondence
mapIdx = [[31, 32], [39, 40], [33, 34], [35, 36], [41, 42], [43, 44], [19, 20], [21, 22], \
          [23, 24], [25, 26], [27, 28], [29, 30], [47, 48], [49, 50], [53, 54], [51, 52], \
          [55, 56], [37, 38], [45, 46]]

# visualize
colors = [[255, 0, 0], [255, 85, 0], [255, 170, 0], [255, 255, 0], [170, 255, 0], [85, 255, 0],
          [0, 255, 0], \
          [0, 255, 85], [0, 255, 170], [0, 255, 255], [0, 170, 255], [0, 85, 255], [0, 0, 255],
          [85, 0, 255], \
          [170, 0, 255], [255, 0, 255], [255, 0, 170], [255, 0, 85]]

ori_paf_idx = [12, 13, 20, 21, 14, 15, 16, 17, 22, 23, 24, 25, 0, 1, 2, 3, \
                4, 5, 6, 7, 8, 9, 10, 11, 28, 29, 30, 31, 34,35, 32, 33, 36, 37, 18, 19, 26, 27]
flip_paf_idx = [20, 21, 12, 13, 22, 23,  24, 25, 14, 15, 16, 17, 6, 7, 8, 9, \
                10, 11, 0, 1, 2, 3, 4, 5, 28, 29, 32, 33, 36, 37, 30, 31, 34,35,26, 27,18, 19]
x_paf_idx = [20, 12, 22, 24, 14, 16, 6, 8, \
                10, 0, 2, 4, 28, 32, 36, 30, 34,26,18]

def recover_flipping_output(oriImg, heatmap_ori_size, paf_ori_size, part_ori_size):

    heatmap_ori_size = heatmap_ori_size[:, ::-1, :]
    heatmap_flip_size = np.zeros((oriImg.shape[0], oriImg.shape[1], 19))
    heatmap_flip_size[:,:,left_part_idx] = heatmap_ori_size[:,:,right_part_idx]
    heatmap_flip_size[:,:,right_part_idx] = heatmap_ori_size[:,:,left_part_idx]
    heatmap_flip_size[:,:,0:2] = heatmap_ori_size[:,:,0:2]

    paf_ori_size = paf_ori_size[:, ::-1, :]
    paf_flip_size = np.zeros((oriImg.shape[0], oriImg.shape[1], 38))
    paf_flip_size[:,:,ori_paf_idx] = paf_ori_size[:,:,flip_paf_idx]
    paf_flip_size[:,:,x_paf_idx] = paf_flip_size[:,:,x_paf_idx]*-1

    part_ori_size = part_ori_size[:, ::-1, :]
    part_flip_size = np.zeros((oriImg.shape[0], oriImg.shape[1], 15))
    part_flip_size[:,:,human_ori_part] = part_ori_size[:,:,human_part]
    return heatmap_flip_size, paf_flip_size, part_flip_size

def recover_flipping_output2(oriImg, part_ori_size):

    part_ori_size = part_ori_size[:, ::-1, :]
    part_flip_size = np.zeros((oriImg.shape[0], oriImg.shape[1], 15))
    part_flip_size[:,:,human_ori_part] = part_ori_size[:,:,human_part]
    return part_flip_size

def part_thresholding(seg_argmax):
    background = 0.6
    head = 0.5
    torso = 0.8

    rightfoot = 0.55 
    leftfoot = 0.55
    leftthigh = 0.55
    rightthigh = 0.55
    leftshank = 0.55
    rightshank = 0.55
    rightupperarm = 0.55
    leftupperarm = 0.55
    rightforearm = 0.55
    leftforearm = 0.55
    lefthand = 0.55
    righthand = 0.55
    
    part_th = [background, head, torso, leftupperarm ,rightupperarm, leftforearm, rightforearm, lefthand, righthand, leftthigh, rightthigh, leftshank, rightshank, leftfoot, rightfoot]
    th_mask = np.zeros(seg_argmax.shape)
    for indx in range(15):
        part_prediction = (seg_argmax==indx)
        part_prediction = part_prediction*part_th[indx]
        th_mask += part_prediction

    return th_mask


def process (oriImg, flipImg, params, model_params, model, Img):
    
    input_scale = 1.0
    multiplier = [x * model_params['boxsize'] / oriImg.shape[0] for x in params['scale_search']]

    heatmap_avg = np.zeros((oriImg.shape[0], oriImg.shape[1], 19))
    paf_avg = np.zeros((oriImg.shape[0], oriImg.shape[1], 38))
    seg_avg = np.zeros((oriImg.shape[0], oriImg.shape[1], 15))

    segmap_scale1 = np.zeros((oriImg.shape[0], oriImg.shape[1], seg_num))
    segmap_scale2 = np.zeros((oriImg.shape[0], oriImg.shape[1], seg_num))
    segmap_scale3 = np.zeros((oriImg.shape[0], oriImg.shape[1], seg_num))
    segmap_scale4 = np.zeros((oriImg.shape[0], oriImg.shape[1], seg_num))

    segmap_scale5 = np.zeros((oriImg.shape[0], oriImg.shape[1], seg_num))
    segmap_scale6 = np.zeros((oriImg.shape[0], oriImg.shape[1], seg_num))
    segmap_scale7 = np.zeros((oriImg.shape[0], oriImg.shape[1], seg_num))
    segmap_scale8 = np.zeros((oriImg.shape[0], oriImg.shape[1], seg_num))

    for m in range(len(multiplier)):
        scale = multiplier[m]*input_scale
        imageToTest = cv2.resize(oriImg, (0, 0), fx=scale, fy=scale, interpolation=cv2.INTER_CUBIC)

        pad = [ 0,
                0, 
                (imageToTest.shape[0] - model_params['stride']) % model_params['stride'],
                (imageToTest.shape[1] - model_params['stride']) % model_params['stride']
              ]
        
        imageToTest_padded = np.pad(imageToTest, ((0, pad[2]), (0, pad[3]), (0, 0)), mode='constant', constant_values=((0, 0), (0, 0), (0, 0)))

        input_img = imageToTest_padded[np.newaxis, ...]
        
        # print( "\tActual size fed into NN: ", input_img.shape)

        output_blobs = model.predict(input_img)

        heatmap = np.squeeze(output_blobs[1])  # output 1 is heatmaps
        heatmap = cv2.resize(heatmap, (0, 0), fx=model_params['stride'], fy=model_params['stride'],
                             interpolation=cv2.INTER_CUBIC)
        heatmap = heatmap[:imageToTest_padded.shape[0] - pad[2], :imageToTest_padded.shape[1] - pad[3],
                  :]
        heatmap = cv2.resize(heatmap, (oriImg.shape[1], oriImg.shape[0]), interpolation=cv2.INTER_CUBIC)

        paf = np.squeeze(output_blobs[0])  # output 0 is PAFs
        paf = cv2.resize(paf, (0, 0), fx=model_params['stride'], fy=model_params['stride'],
                         interpolation=cv2.INTER_CUBIC)
        paf = paf[:imageToTest_padded.shape[0] - pad[2], :imageToTest_padded.shape[1] - pad[3], :]
        paf = cv2.resize(paf, (oriImg.shape[1], oriImg.shape[0]), interpolation=cv2.INTER_CUBIC)

        seg = np.squeeze(output_blobs[2])
        seg = cv2.resize(seg, (0, 0), fx=model_params['stride'], fy=model_params['stride'],
                             interpolation=cv2.INTER_CUBIC)
        seg = seg[:imageToTest_padded.shape[0] - pad[2], :imageToTest_padded.shape[1] - pad[3], :]
        seg = cv2.resize(seg, (oriImg.shape[1], oriImg.shape[0]), interpolation=cv2.INTER_CUBIC)

        if m==0:
            segmap_scale1 = seg
        elif m==1:
            segmap_scale2 = seg         
        elif m==2:
            segmap_scale3 = seg
        elif m==3:
            segmap_scale4 = seg


    # flipping
    for m in range(len(multiplier)):
        scale = multiplier[m]
        imageToTest = cv2.resize(flipImg, (0, 0), fx=scale, fy=scale, interpolation=cv2.INTER_CUBIC)
        pad = [ 0,
                0, 
                (imageToTest.shape[0] - model_params['stride']) % model_params['stride'],
                (imageToTest.shape[1] - model_params['stride']) % model_params['stride']
              ]
        
        imageToTest_padded = np.pad(imageToTest, ((0, pad[2]), (0, pad[3]), (0, 0)), mode='constant', constant_values=((0, 0), (0, 0), (0, 0)))
        input_img = imageToTest_padded[np.newaxis, ...]
        # print( "\tActual size fed into NN: ", input_img.shape)
        output_blobs = model.predict(input_img)

        # extract outputs, resize, and remove padding
        heatmap = np.squeeze(output_blobs[1])  # output 1 is heatmaps
        heatmap = cv2.resize(heatmap, (0, 0), fx=model_params['stride'], fy=model_params['stride'],interpolation=cv2.INTER_CUBIC)
        heatmap = heatmap[:imageToTest_padded.shape[0] - pad[2], :imageToTest_padded.shape[1] - pad[3], :]
        heatmap = cv2.resize(heatmap, (oriImg.shape[1], oriImg.shape[0]), interpolation=cv2.INTER_CUBIC)

        paf = np.squeeze(output_blobs[0])  # output 0 is PAFs
        paf = cv2.resize(paf, (0, 0), fx=model_params['stride'], fy=model_params['stride'],interpolation=cv2.INTER_CUBIC)
        paf = paf[:imageToTest_padded.shape[0] - pad[2], :imageToTest_padded.shape[1] - pad[3], :]
        paf = cv2.resize(paf, (oriImg.shape[1], oriImg.shape[0]), interpolation=cv2.INTER_CUBIC)

        seg = np.squeeze(output_blobs[2])
        seg = cv2.resize(seg, (0, 0), fx=model_params['stride'], fy=model_params['stride'],
                             interpolation=cv2.INTER_CUBIC)
        seg = seg[:imageToTest_padded.shape[0] - pad[2], :imageToTest_padded.shape[1] - pad[3], :]
        seg = cv2.resize(seg, (oriImg.shape[1], oriImg.shape[0]), interpolation=cv2.INTER_CUBIC)

        heatmap_recover, paf_recover, seg_recover = recover_flipping_output(oriImg, heatmap, paf, seg)

        heatmap_avg = heatmap_avg + heatmap_recover
        paf_avg = paf_avg + paf_recover

        if m==0:
            segmap_scale5 = seg_recover
        elif m==1:
            segmap_scale6 = seg_recover         
        elif m==2:
            segmap_scale7 = seg_recover
        elif m==3:
            segmap_scale8 = seg_recover

    heatmap_avg = heatmap_avg / (len(multiplier)*2)
    paf_avg = paf_avg / (len(multiplier)*2)

    segmap_a = np.maximum(segmap_scale1,segmap_scale2)
    segmap_b = np.maximum(segmap_scale4,segmap_scale3)
    segmap_c = np.maximum(segmap_scale5,segmap_scale6)
    segmap_d = np.maximum(segmap_scale7,segmap_scale8)
    seg_ori = np.maximum(segmap_a, segmap_b)
    seg_flip = np.maximum(segmap_c, segmap_d)
    seg_avg = np.maximum(seg_ori, seg_flip)

    all_peaks = []
    peak_counter = 0

    for part in range(18):
        map_ori = heatmap_avg[:, :, part]
        map = gaussian_filter(map_ori, sigma=3)

        map_left = np.zeros(map.shape)
        map_left[1:, :] = map[:-1, :]
        map_right = np.zeros(map.shape)
        map_right[:-1, :] = map[1:, :]
        map_up = np.zeros(map.shape)
        map_up[:, 1:] = map[:, :-1]
        map_down = np.zeros(map.shape)
        map_down[:, :-1] = map[:, 1:]

        peaks_binary = np.logical_and.reduce(
            (map >= map_left, map >= map_right, map >= map_up, map >= map_down, map > params['thre1']))
        peaks = list(zip(np.nonzero(peaks_binary)[1], np.nonzero(peaks_binary)[0]))  # note reverse
        peaks_with_score = [x + (map_ori[x[1], x[0]],) for x in peaks]
        id = range(peak_counter, peak_counter + len(peaks))
        peaks_with_score_and_id = [peaks_with_score[i] + (id[i],) for i in range(len(id))]

        all_peaks.append(peaks_with_score_and_id)
        peak_counter += len(peaks)

    connection_all = []
    special_k = []
    mid_num = 10

    for k in range(len(mapIdx)):
        score_mid = paf_avg[:, :, [x - 19 for x in mapIdx[k]]]
        candA = all_peaks[limbSeq[k][0] - 1]
        candB = all_peaks[limbSeq[k][1] - 1]
        nA = len(candA)
        nB = len(candB)
        indexA, indexB = limbSeq[k]
        if (nA != 0 and nB != 0):
            connection_candidate = []
            for i in range(nA):
                for j in range(nB):
                    vec = np.subtract(candB[j][:2], candA[i][:2])
                    norm = math.sqrt(vec[0] * vec[0] + vec[1] * vec[1])
                    # failure case when 2 body parts overlaps
                    if norm == 0:
                        continue
                    vec = np.divide(vec, norm)

                    startend = list(zip(np.linspace(candA[i][0], candB[j][0], num=mid_num), \
                                   np.linspace(candA[i][1], candB[j][1], num=mid_num)))

                    vec_x = np.array(
                        [score_mid[int(round(startend[I][1])), int(round(startend[I][0])), 0] \
                         for I in range(len(startend))])
                    vec_y = np.array(
                        [score_mid[int(round(startend[I][1])), int(round(startend[I][0])), 1] \
                         for I in range(len(startend))])

                    score_midpts = np.multiply(vec_x, vec[0]) + np.multiply(vec_y, vec[1])
                    score_with_dist_prior = sum(score_midpts) / len(score_midpts) + min(
                        0.5 * oriImg.shape[0] / norm - 1, 0)
                    criterion1 = len(np.nonzero(score_midpts > params['thre2'])[0]) > 0.8 * len(
                        score_midpts)
                    criterion2 = score_with_dist_prior > 0
                    if criterion1 and criterion2:
                        connection_candidate.append([i, j, score_with_dist_prior,
                                                     score_with_dist_prior + candA[i][2] + candB[j][2]])

            connection_candidate = sorted(connection_candidate, key=lambda x: x[2], reverse=True)
            connection = np.zeros((0, 5))
            for c in range(len(connection_candidate)):
                i, j, s = connection_candidate[c][0:3]
                if (i not in connection[:, 3] and j not in connection[:, 4]):
                    connection = np.vstack([connection, [candA[i][3], candB[j][3], s, i, j]])
                    if (len(connection) >= min(nA, nB)):
                        break

            connection_all.append(connection)
        else:
            special_k.append(k)
            connection_all.append([])

    # last number in each row is the total parts number of that person
    # the second last number in each row is the score of the overall configuration
    subset = -1 * np.ones((0, 20))
    candidate = np.array([item for sublist in all_peaks for item in sublist])

    for k in range(len(mapIdx)):
        if k not in special_k:
            partAs = connection_all[k][:, 0]
            partBs = connection_all[k][:, 1]
            indexA, indexB = np.array(limbSeq[k]) - 1

            for i in range(len(connection_all[k])):  # = 1:size(temp,1)
                found = 0
                subset_idx = [-1, -1]
                for j in range(len(subset)):  # 1:size(subset,1):
                    if subset[j][indexA] == partAs[i] or subset[j][indexB] == partBs[i]:
                        subset_idx[found] = j
                        found += 1

                if found == 1:
                    j = subset_idx[0]
                    if (subset[j][indexB] != partBs[i]):
                        subset[j][indexB] = partBs[i]
                        subset[j][-1] += 1
                        subset[j][-2] += candidate[partBs[i].astype(int), 2] + connection_all[k][i][2]
                elif found == 2:  # if found 2 and disjoint, merge them
                    j1, j2 = subset_idx
                    membership = ((subset[j1] >= 0).astype(int) + (subset[j2] >= 0).astype(int))[:-2]
                    if len(np.nonzero(membership == 2)[0]) == 0:  # merge
                        subset[j1][:-2] += (subset[j2][:-2] + 1)
                        subset[j1][-2:] += subset[j2][-2:]
                        subset[j1][-2] += connection_all[k][i][2]
                        subset = np.delete(subset, j2, 0)
                    else:  # as like found == 1
                        subset[j1][indexB] = partBs[i]
                        subset[j1][-1] += 1
                        subset[j1][-2] += candidate[partBs[i].astype(int), 2] + connection_all[k][i][2]
                # if find no partA in the subset, create a new subset
                elif not found and k < 17:
                    row = -1 * np.ones(20)
                    row[indexA] = partAs[i]
                    row[indexB] = partBs[i]
                    row[-1] = 2
                    row[-2] = sum(candidate[connection_all[k][i, :2].astype(int), 2]) + \
                              connection_all[k][i][2]
                    subset = np.vstack([subset, row])

    # delete some rows of subset which has few parts occur
    deleteIdx = []
    for i in range(len(subset)):
        if subset[i][-1] < 4 or subset[i][-2] / subset[i][-1] < 0.4:
            deleteIdx.append(i)
    subset = np.delete(subset, deleteIdx, axis=0)

    # canvas = cv2.imread(input_image)
    '''
    for i in range(18):
        for j in range(len(all_peaks[i])):
            cv2.circle(canvas, all_peaks[i][j][0:2], 4, colors[i], thickness=-1)
    '''
    stickwidth = 4

    for i in range(17):
        for n in range(len(subset)):
            index = subset[n][np.array(limbSeq[i]) - 1]
            if -1 in index:
                continue
            cur_canvas = Img.copy()
            Y = candidate[index.astype(int), 0]
            X = candidate[index.astype(int), 1]
            mX = np.mean(X)
            mY = np.mean(Y)
            length = ((X[0] - X[1]) ** 2 + (Y[0] - Y[1]) ** 2) ** 0.5
            angle = math.degrees(math.atan2(X[0] - X[1], Y[0] - Y[1]))
            polygon = cv2.ellipse2Poly((int(mY), int(mX)), (int(length / 2), stickwidth), int(angle), 0,
                                       360, 1)
            cv2.fillConvexPoly(Img, polygon, colors[i])
            canvas = cv2.addWeighted(Img, 0.3, cur_canvas, 0.7, 0)
    people = []
    subset_int = subset.astype(int)
    for i in range(len(subset_int)):
        people.append([])
        for j in range(18):
            if subset_int[i][j] == -1:
                people[-1].append([-1, -1])
            else:
                people[-1].append([candidate[subset_int[i][j], 0], candidate[subset_int[i][j], 1]])

    # code.interact(local=locals())
    return canvas, heatmap_avg, paf_avg, people, seg_avg


def run_image(input_path:str, output_path:str):

    print('-' * 20)
    print('Downloading Weights File...\nPlease Wait...')
    print('-' * 20)
    weights_path = tf.keras.utils.get_file('cdcl.h5',
              					'https://github.com/Adeel-Intizar/Xtreme-Vision/releases/download/1.0/cdcl_model.h5',
                  			cache_subdir = 'weights/', cache_dir = 'xtreme_vision')    

    print('Processing Started...')
    # load model
    model = get_testing_model_resnet101() 
    model.load_weights(weights_path)
    params, model_params = config_reader()

    scale_list = [1.0]
    params['scale_search'] = scale_list

    if input_path.endswith(".png") or input_path.endswith(".jpg"):

        Img = cv2.imread(input_path)
        flipImg = cv2.flip(Img, 1)
        oriImg = (Img / 256.0) - 0.5
        flipImg = (flipImg / 256.0) - 0.5

        canvas, heatmap, paf, people, seg = process(oriImg, flipImg, params, model_params, model, Img)
        print('Done... Processing is finished.')
        cv2.imwrite(output_path, canvas)
        # seg_argmax = np.argmax(seg, axis=-1)
        # seg_max = np.max(seg, axis=-1)
        # th_mask = part_thresholding(seg_argmax)
        # seg_max_thres = (seg_max > 0.1).astype(np.uint8)
        # seg_argmax *= seg_max_thres
        # seg_canvas = human_seg_combine_argmax(seg_argmax)
        # # cur_canvas = cv2.imread(args.input_folder+'/'+filename)
        # canvas = cv2.addWeighted(seg_canvas, 0.6, Img, 0.4, 0)
        # # filename = '%s/%s.jpg'%(args.output_folder,'seg_'+filename)
        # cv2.imwrite('seg-' + output_path, canvas) 


def run_video(input_path:str, output_path:str, fps:int = 25):

    print('-' * 20)
    print('Downloading Weights File...\nPlease Wait...')
    print('-' * 20)
    weights_path = tf.keras.utils.get_file('cdcl.h5',
              					'https://github.com/Adeel-Intizar/Xtreme-Vision/releases/download/1.0/cdcl_model.h5',
                  			cache_subdir = 'weights/', cache_dir = 'xtreme_vision')    

    
    
    # load model
    model = get_testing_model_resnet101() 
    model.load_weights(weights_path)
    params, model_params = config_reader()
    params['scale_search'] = [1.0]

    if not (input_path.endswith(".mp4") or input_path.endswith(".avi")):
        raise RuntimeError ('Invalid video format, Valid formats are mp4 and avi.')
    else:

        out = None
        cap = cv2.VideoCapture(input_path)
        length = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
        print(f'\nThere are {length} Frames in this video')
        print('-' * 20)
        print('Detecting Objects in the Video... Please Wait...')
        print('-' * 20)

        while(cap.isOpened()):
            retreive, frame = cap.read()
            if not retreive:
                break

            flipImg = cv2.flip(frame, 1)
            oriImg = (frame / 256.0) - 0.5
            flipImg = (flipImg / 256.0) - 0.5
            

            canvas, heatmap, paf, people, seg = process(oriImg, flipImg, params, model_params, model, frame)

            if out is None:
                fourcc = cv2.VideoWriter_fourcc(*'DIVX')
                out = cv2.VideoWriter(output_path, fourcc, fps, (frame.shape[1], frame.shape[0]))
            
            out.write(canvas)
        print('Done. Processing has been Finished... Please Check Output Video.')
        out.release()
        cap.release()


"""
MIT License

Copyright (c) 2020 Adeel <kingadeel2017@outlook.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""



import cv2
import numpy as np
import warnings
import os
warnings.filterwarnings('ignore')

from xtreme_vision.Detection.retinanet import models
from xtreme_vision.Detection.retinanet.utils.image import preprocess_image, resize_image
from xtreme_vision.Detection.retinanet.utils.visualization import draw_box, draw_caption
from xtreme_vision.Detection.retinanet.utils.colors import label_color


class RetinaNet:
    def __init__(self):
        self.model = None
        self.weights_path = ""
        self.output_path = ""
        self.min_prob = 0
        
        self.classes = {0: 'person', 1: 'bicycle', 2: 'car', 3: 'motorcycle', 4: 'airplane', 5: 'bus', 6: 'train', 7: 'truck', 8: 'boat',
                   9: 'traffic light', 10: 'fire hydrant', 11: 'stop sign', 12: 'parking meter', 13: 'bench', 14: 'bird', 15: 'cat', 16: 'dog',
                   17: 'horse', 18: 'sheep', 19: 'cow', 20: 'elephant', 21: 'bear', 22: 'zebra', 23: 'giraffe', 24: 'backpack', 25: 'umbrella',
                   26: 'handbag', 27: 'tie', 28: 'suitcase', 29: 'frisbee', 30: 'skis', 31: 'snowboard', 32: 'sports ball', 33: 'kite',
                   34: 'baseball bat', 35: 'baseball glove', 36: 'skateboard', 37: 'surfboard', 38: 'tennis racket', 39: 'bottle', 40: 'wine glass',
                   41: 'cup', 42: 'fork', 43: 'knife', 44: 'spoon', 45: 'bowl', 46: 'banana', 47: 'apple', 48: 'sandwich', 49: 'orange',
                   50: 'broccoli', 51: 'carrot', 52: 'hot dog', 53: 'pizza', 54: 'donut', 55: 'cake', 56: 'chair', 57: 'couch', 58: 'potted plant',
                   59: 'bed', 60: 'dining table', 61: 'toilet', 62: 'tv', 63: 'laptop', 64: 'mouse', 65: 'remote', 66: 'keyboard', 67: 'cell phone',
                   68: 'microwave', 69: 'oven', 70: 'toaster', 71: 'sink', 72: 'refrigerator', 73: 'book', 74: 'clock', 75: 'vase', 76: 'scissors',
                   77: 'teddy bear', 78: 'hair drier', 79: 'toothbrush'}
        
    def load_model(self, Weights_Path:str = None, classes:dict=None, backbone:str='resnet50'):
        
        if classes != None:
            self.classes = classes
        
        if Weights_Path is None:
            raise RuntimeError ('Weights_Path should not be None.')
        self.weights_path = Weights_Path
        self.model = models.load_model(self.weights_path, backbone_name = backbone)
        
    def predict(self, img_path: np.ndarray = None, output_path:str = None, debug=True, custom_objects = None,
                extract_objects = False, min_prob:float = 0.25):
        
        if img_path is None:
            raise RuntimeError ('img_path should not be None.')
        
        image = img_path
        self.output_path = output_path
        self.min_prob = min_prob
        self.custom_objects = custom_objects

        draw = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        copy = draw.copy()
        image = preprocess_image(image)
        image, scale = resize_image(image)
        boxes, scores, labels = self.model.predict_on_batch(np.expand_dims(image, axis=0))
        boxes /= scale


        for index, (box, score, label) in enumerate(zip(boxes[0], scores[0], labels[0])):
  
            if score < self.min_prob:
              continue


            elif (custom_objects != None):
              check_name = self.classes.get(label, 'invalid')
              check = custom_objects.get(check_name, 'invalid')
              if (check == "invalid"):
                continue


            if extract_objects:
              detection_details = boxes[0, index, :4].astype(int)
              splitted_copy = copy.copy()[detection_details[1]:detection_details[3],
                                            detection_details[0]:detection_details[2]]

              directory = './extracted_objects'
              if (os.path.exists(directory) == False):
                os.mkdir(directory)
              
              splitted_image_path = os.path.join(directory, str(self.classes[label]) + "-" + str(index) + ".jpg")
              cv2.imwrite(splitted_image_path, splitted_copy)
                
            color = label_color(label)
            b = box.astype(int)
            draw_box(draw, b, color=color)
            caption = "{} {:.0f}%".format(self.classes[label], (score * 100))
            draw_caption(draw, b, caption)
        detected_img =cv2.cvtColor(draw, cv2.COLOR_RGB2BGR)
        
        if debug:
            cv2.imwrite(self.output_path, detected_img)
        
        return detected_img



import cv2
import numpy as np

#part_ids = [ 0,  2,  4,  5,  6,  8, 13, 14, 16, 17, 18, 19, 20, 21, 24]
# the order is: (left right flipped)
# background, head, torso, left upper arm ,right upper arm, left forearm, right forearm,
#  left hand, right hand, left thigh, right thigh, left shank, right shank, left foot, right foot
part_ids = [0, 13, 2, 5, 8, 19, 20, 4, 24, 18, 6, 21, 16, 14, 17]

png_idx = [0, 14, 1, 11, 10, 13, 12, 2, 3, 6, 7, 8, 9, 5, 4]


def human_seg_spread_channel(human_seg_map):
    x = human_seg_map // 127
    x = x * np.array([9, 3, 1])
    x = np.add.reduce(x, 2)
    res = []
    for i in part_ids:
        res.append((x == i))
    res = np.stack(res, axis=-1)
    return res.astype(np.float32)

def human_seg_combine_channel(human_seg_split_map):
    segmap = np.add.reduce(human_seg_split_map * np.array(png_idx), 2)
    return np.stack([segmap], axis=-1).astype(np.uint8)

def human_seg_combine_argmax(human_seg_argmax_map):
    onehot = np.stack([(human_seg_argmax_map == i).astype(np.uint8) for i in range(15)], axis=-1)
    return human_seg_combine_channel(onehot)


"""
MIT License

Copyright (c) 2020 Hyeonki Hong <hhk7734@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from tensorflow.keras import Model

from xtreme_vision.Detection.yolov4.model.backbone import CSPDarknet53, CSPDarknet53Tiny
from xtreme_vision.Detection.yolov4.model.head import YOLOv3Head, YOLOv3HeadTiny
from xtreme_vision.Detection.yolov4.model.neck import PANet, PANetTiny


class YOLOv4(Model):
    """
    Path Aggregation Network(PAN)
    Spatial Attention Module(SAM)
    Bounding Box(BBox)
    """

    def __init__(
        self,
        anchors,
        num_classes: int,
        xyscales,
        activation0: str = "mish",
        activation1: str = "leaky",
        kernel_regularizer=None,
    ):
        super(YOLOv4, self).__init__(name="YOLOv4")
        self.csp_darknet53 = CSPDarknet53(
            activation0=activation0,
            activation1=activation1,
            kernel_regularizer=kernel_regularizer,
        )
        self.panet = PANet(
            num_classes=num_classes,
            activation=activation1,
            kernel_regularizer=kernel_regularizer,
        )
        self.yolov3_head = YOLOv3Head(
            anchors=anchors, num_classes=num_classes, xysclaes=xyscales
        )

    def call(self, x):
        """
        @param x: Dim(batch, input_size, input_size, channels)
            The element has a value between 0.0 and 1.0.

        @return (s_pred, m_pred, l_pred)
            downsampling_size = [8, 16, 32]
            output_size = input_size // downsampling_szie
            s_pred = Dim(batch, output_size[0], output_size[0],
                                        num_anchors, (5 + num_classes))
            l_pred = Dim(batch, output_size[1], output_size[1],
                                        num_anchors, (5 + num_classes))
            m_pred = Dim(batch, output_size[2], output_size[2],
                                        num_anchors, (5 + num_classes))

        Ref: https://arxiv.org/abs/1612.08242 - YOLOv2

        5 + num_classes = (t_x, t_y, t_w, t_h, t_o, c_0, c_1, ...)
                            => (b_x, b_y, b_w, b_h, conf, prob_0, prob_1, ...)

        A top-left coordinate of a grid is (c_x, c_y).
        A dimension prior is (p_w, p_h).(== anchor size)
                [[[12, 16],   [19, 36],   [40, 28]  ],
                 [[36, 75],   [76, 55],   [72, 146] ],
                 [[142, 110], [192, 243], [459, 401]]]
        Pr == Probability.

        b_x = sigmoid(t_x) + c_x
        b_y = sigmoid(t_y) + c_y
        b_w = p_w * exp(t_w)
        b_h = p_h * exp(t_h)
        sigmoid(t_o) == confidence == Pr(Object) ∗ IoU(b, Object)
        sigmoid(c_i) == conditional class probability == Pr(Class_i|Object)
        """
        x = self.csp_darknet53(x)
        x = self.panet(x)
        x = self.yolov3_head(x)
        return x


class YOLOv4Tiny(Model):
    def __init__(
        self,
        anchors,
        num_classes: int,
        xyscales,
        activation: str = "leaky",
        kernel_regularizer=None,
    ):
        super(YOLOv4Tiny, self).__init__(name="YOLOv4Tiny")
        self.csp_darknet53_tiny = CSPDarknet53Tiny(
            activation=activation,
            kernel_regularizer=kernel_regularizer,
        )
        self.panet_tiny = PANetTiny(
            num_classes=num_classes,
            activation=activation,
            kernel_regularizer=kernel_regularizer,
        )
        self.yolov3_head_tiny = YOLOv3HeadTiny(
            anchors=anchors, num_classes=num_classes, xysclaes=xyscales
        )

    def call(self, x):
        x = self.csp_darknet53_tiny(x)
        x = self.panet_tiny(x)
        x = self.yolov3_head_tiny(x)
        return x



import tensorflow as tf
import numpy as np
from PIL import Image
from .deeplab import Deeplab_xcep_pascal
from .semantic import obtain_segmentation
import cv2
import time


class alter_bg():
  INPUT_TENSOR_NAME = 'ImageTensor:0'
  OUTPUT_TENSOR_NAME = 'SemanticPredictions:0'
  INPUT_SIZE = 513
  FROZEN_GRAPH_NAME = 'frozen_inference_graph'

  def __init__(self, model_type = "h5"):
    global model_file
    self.model_type = model_type
    model_file = model_type

    self.model = Deeplab_xcep_pascal()

    
  def load_pascalvoc_model(self, model_path):
    if model_file == "pb":
      self.graph = tf.Graph()

      graph_def = None

      with tf.compat.v1.gfile.GFile(model_path, 'rb')as file_handle:
        graph_def = tf.compat.v1.GraphDef.FromString(file_handle.read())

      if graph_def is None:
        raise RuntimeError('Cannot find inference graph')

      with self.graph.as_default():
        tf.graph_util.import_graph_def(graph_def, name='')

      self.sess = tf.compat.v1.Session(graph=self.graph)

    else:   
      self.model.load_weights(model_path)

  def segmentAsPascalvoc(self, image_path, process_frame = False):
    if model_file == "pb":

      if process_frame == True:
        image = image_path
      else:
        image = cv2.imread(image_path)

      h, w, n = image.shape
     
      if n > 3:
        image = cv2.cvtColor(image, cv2.COLOR_BGRA2BGR)
      resize_ratio = 1.0 * self.INPUT_SIZE / max(w, h)
      target_size = (int(resize_ratio * w), int(resize_ratio * h))
      image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
      resized_image = cv2.resize(image, target_size, interpolation=cv2.INTER_AREA)
      batch_seg_map = self.sess.run(
        self.OUTPUT_TENSOR_NAME,
        feed_dict={self.INPUT_TENSOR_NAME: [np.asarray(resized_image)]})
     
      seg_image = batch_seg_map[0]
      raw_labels = seg_image
      labels = obtain_segmentation(seg_image)
      labels = np.array(Image.fromarray(labels.astype('uint8')).resize((w, h)))
      labels = cv2.cvtColor(labels, cv2.COLOR_RGB2BGR)

      return raw_labels, labels 

    else:               
      trained_image_width=512
      mean_subtraction_value=127.5

      if process_frame == True:
        image = image_path

      else:  
        image = np.array(Image.open(image_path))     
   

      # resize to max dimension of images from training dataset
      w, h, n = image.shape

      if n > 3:
        image = cv2.cvtColor(image, cv2.COLOR_BGRA2RGB)
    

      ratio = float(trained_image_width) / np.max([w, h])
      resized_image = np.array(Image.fromarray(image.astype('uint8')).resize((int(ratio * h), int(ratio * w))))
      resized_image = (resized_image / mean_subtraction_value) -1


      # pad array to square image to match training images
      pad_x = int(trained_image_width - resized_image.shape[0])
      pad_y = int(trained_image_width - resized_image.shape[1])
      resized_image = np.pad(resized_image, ((0, pad_x), (0, pad_y), (0, 0)), mode='constant')

      #run prediction
      res = self.model.predict(np.expand_dims(resized_image, 0))
    
      labels = np.argmax(res.squeeze(), -1)
      # remove padding and resize back to original image
      if pad_x > 0:
        labels = labels[:-pad_x]
      if pad_y > 0:
        labels = labels[:, :-pad_y]

      raw_labels = labels
        
      #Apply segmentation color map
      labels = obtain_segmentation(labels)
      labels = np.array(Image.fromarray(labels.astype('uint8')).resize((h, w)))
    
    
      new_img = cv2.cvtColor(labels, cv2.COLOR_RGB2BGR)


      return raw_labels, new_img 



  ####FUNCTION TO FILTER OUT OTHER DETECTIONS AND ENSURE THE DETECTION OF A DISTINCT OBJECT####


  def target_obj(self, segment):

    self.segment = segment
    if segment == "person":
      segment = (255,255,255)
      return segment

   
    if segment == "car":
      segment = (128,128,128)  
      return segment
    
    if segment == "aeroplane":
      segment = (128,0,0) 
      return segment

    
    if segment == "bicycle": 
      segment = (0,128,0) 
      return segment

     
    if segment == "bird":
      segment = (255, 255, 255)
      return segment

     
    if segment == "boat":
      segment = (0,0,128)
      return segment

    
    if segment == "bottle":
      segment = (128,0,128)
      return segment
    
    
    if segment == "bus":
      segment = (0,128,128) 
      return segment

    
    if segment == "cat":
      segment = (64,0,0) 
      return segment

    
    if segment == "chair":  
      segment = (192,0,0)
      return segment
      
    
    if segment == "cow":
      segment = (64,128,0)
      return segment

    
    if segment == "diningtable": 
      segment = (192,128,0) 
      return segment

    if segment == "dog":
      segment = (64,0,128) 
      return segment

    
    if segment == "horse":  
      segment = (192,0,128)
      return segment

    
    if segment == "motorbike":  
      segment = (64, 128, 128) 
      return segment
    

    
    if segment == "pottedplant":
      segment = (0,64,0)
      return segment

    
    if segment == "sheep":
      segment = (128,64,0)
      return segment 
      
    
    if segment == "sofa":
      segment = (0,192,0) 
      return segment
      
    
    if segment == "train":
      segment = (128,192, 0) 
      return segment
      
    
    if segment == "monitor":
      segment = (0,64,128)  
      return segment


  #### ALTER IMAGE BACKGROUND WITH A NEW PICTURE ###

  def change_bg_img(self, f_image_path,b_image_path, output_image_name = None, verbose = None, detect = None):
    if verbose is not None:
      print("processing image......")

    seg_image = self.segmentAsPascalvoc(f_image_path)
    
    if detect is not None:
      target_class = self.target_obj(detect)
      seg_image[1][seg_image[1] != target_class] = 0
      
    ori_img = cv2.imread(f_image_path)
    
    bg_img = cv2.imread(b_image_path)
    w, h, _ = ori_img.shape
    bg_img = cv2.resize(bg_img, (h,w))

    result = np.where(seg_image[1], ori_img, bg_img)
    if output_image_name is not None:
      cv2.imwrite(output_image_name, result)

    return result


  ##### GIVE THE BACKGROUND OF AN IMAGE A DISTINCT COLOR ######
    
  def color_bg(self, image_path, colors, output_image_name = None, verbose = None, detect = None):
    if verbose is not None:
      print("processing image......")
      
    seg_image = self.segmentAsPascalvoc(image_path)
    if detect is not None:
      target_class = self.target_obj(detect)
      seg_image[1][seg_image[1] != target_class] = 0

    ori_img = cv2.imread(image_path)
    
    obtain_img = cv2.subtract(seg_image[1], ori_img)
    out = cv2.subtract(seg_image[1], obtain_img)

    out[np.where((out == [0, 0, 0]).all(axis = 2))] = [colors]
    out = cv2.cvtColor(out, cv2.COLOR_BGR2RGB)
    result = np.where(seg_image[1], ori_img, out)


    if output_image_name is not None:
      cv2.imwrite(output_image_name, result)

    return result


   
   ##### GIVE THE BACKGROUND OF A FRAME A DISTINCT COLOR ######

  def color_frame(self, frame, colors, verbose = None, detect = None):
    if verbose is not None:
      print("processing frame....")

    seg_frame = self.segmentAsPascalvoc(frame, process_frame=True)

    if detect is not None:
      target_class = self.target_obj(detect)
      seg_frame[1][seg_frame[1] != target_class] = 0
    
    obtain_frame = cv2.subtract(seg_frame[1], frame)
    out = cv2.subtract(seg_frame[1], obtain_frame)
    out[np.where((out == [0, 0, 0]).all(axis = 2))] = [colors]
    out = cv2.cvtColor(out, cv2.COLOR_BGR2RGB)
    result = np.where(seg_frame[1], frame, out)


    return result




  ##### GIVE THE BACKGROUND OF A VIDEO A DISTINCT COLOR ######


  def color_video(self, video_path, colors, frames_per_second = None, output_video_name = None, detect = None):
    capture = cv2.VideoCapture(video_path)
    width = int(capture.get(cv2.CAP_PROP_FRAME_WIDTH))
    height = int(capture.get(cv2.CAP_PROP_FRAME_HEIGHT))
    if frames_per_second is not None:
      save_video = cv2.VideoWriter(output_video_name, cv2.VideoWriter_fourcc(*'DIVX'),frames_per_second, (width, height))
    counter = 0
    start = time.time() 
     
    while True:
        counter += 1
        ret, frame = capture.read()
        if ret:
          seg_frame = self.segmentAsPascalvoc(frame, process_frame=True)

          if detect is not None:
            target_class = self.target_obj(detect)
            seg_frame[1][seg_frame[1] != target_class] = 0

          print("No. of frames:", counter)
          obtain_frame = cv2.subtract(seg_frame[1], frame)
          out = cv2.subtract(seg_frame[1], obtain_frame)

          out[np.where((out == [0, 0, 0]).all(axis = 2))] = [colors]
          out = cv2.cvtColor(out, cv2.COLOR_BGR2RGB)
          out = np.where(seg_frame[1], frame, out)
          
            
          output = cv2.resize(out, (width,height), interpolation=cv2.INTER_AREA)
          if output_video_name is not None:
            save_video.write(output)

        else:

          break

    capture.release()

    end = time.time()
    print(f"Processed {counter} frames in {end-start:.1f} seconds")
      
    if frames_per_second is not None:
        save_video.release()

    return  output



  ##### GIVE THE BACKGROUND OF A CAMERA FRAME A DISTINCT COLOR ######  

  def color_camera(self, cam, colors, frames_per_second = None, check_fps = False,show_frames = False, 
  frame_name = None, verbose = None, output_video_name = None, detect = None):
    capture = cam
    width = int(capture.get(cv2.CAP_PROP_FRAME_WIDTH))
    height = int(capture.get(cv2.CAP_PROP_FRAME_HEIGHT))
    if frames_per_second is not None:
      save_video = cv2.VideoWriter(output_video_name, cv2.VideoWriter_fourcc(*'DIVX'),frames_per_second, (width, height))
    counter = 0
    start = time.time() 
     
    while True:
        counter += 1
        ret, frame = capture.read()
        if ret:
          seg_frame = self.segmentAsPascalvoc(frame, process_frame=True)
          if detect is not None:
            target_class = self.target_obj(detect)
            seg_frame[1][seg_frame[1] != target_class] = 0

          if verbose is not None:
            print("No.of frames:", counter)
          obtain_frame = cv2.subtract(seg_frame[1], frame)
          out = cv2.subtract(seg_frame[1], obtain_frame)

          out[np.where((out == [0, 0, 0]).all(axis = 2))] = [colors]
          out = cv2.cvtColor(out, cv2.COLOR_BGR2RGB)
          out = np.where(seg_frame[1], frame, out)

            
          output = cv2.resize(out, (width,height), interpolation=cv2.INTER_AREA)
          if show_frames == True:
            if frame_name is not None:
              cv2.imshow(frame_name, output)
              if cv2.waitKey(25) & 0xFF == ord('q'):
                break 
          if output_video_name is not None:
            save_video.write(output)

        else:

          break

    if check_fps == True:
        out = capture.get(cv2.CAP_PROP_FPS)
        print(f"{out} frames per seconds") 

    capture.release()

    end = time.time()
    if verbose is not None:
      print(f"Processed {counter} frames in {end-start:.1f} seconds")
      
    if frames_per_second is not None:
      save_video.release()

    return  output


  ##### BLUR THE BACKGROUND OF AN IMAGE #####

  def blur_bg(self, image_path,low = False, moderate = False, extreme = False, output_image_name = None, verbose = None, detect = None):
    if verbose is not None:
      print("processing image......")
      
    seg_image = self.segmentAsPascalvoc(image_path)

    if detect is not None:
      target_class = self.target_obj(detect)
      seg_image[1][seg_image[1] != target_class] = 0
    
    ori_img = cv2.imread(image_path)

    if low == True:
        blur_img = cv2.blur(ori_img, (21,21), 0)

    if moderate == True:
        blur_img = cv2.blur(ori_img, (39,39), 0)

    if extreme == True:
        blur_img = cv2.blur(ori_img, (81,81), 0)

    out = np.where(seg_image[1], ori_img, blur_img)
    
    if output_image_name is not None:
        cv2.imwrite(output_image_name, out)

    return out    



  ##### BLUR THE BACKGROUND OF A FRAME #####

  def blur_frame(self, frame,low = False, moderate = False, extreme = False, verbose = None, detect = None):
    if verbose is not None:
      print("processing frame......")
      
    seg_frame = self.segmentAsPascalvoc(frame, process_frame=True)
    if detect is not None:
      target_class = self.target_obj(detect)
      seg_frame[1][seg_frame[1] != target_class] = 0

    if low == True:
        blur_frame = cv2.blur(frame, (21,21), 0)

    if moderate == True:
        blur_frame = cv2.blur(frame, (39,39), 0)

    if extreme == True:
        blur_frame = cv2.blur(frame, (81,81), 0)

    result = np.where(seg_frame[1], frame, blur_frame)

    return result     


  ####  BLUR THE BACKGROUND OF A VIDEO #####

  def blur_video(self, video_path, low = False, moderate = False, extreme = False, frames_per_second = None,
  output_video_name = None, detect = None):
    capture = cv2.VideoCapture(video_path)
    width = int(capture.get(cv2.CAP_PROP_FRAME_WIDTH))
    height = int(capture.get(cv2.CAP_PROP_FRAME_HEIGHT))
    if frames_per_second is not None:
      save_video = cv2.VideoWriter(output_video_name, cv2.VideoWriter_fourcc(*'DIVX'),frames_per_second, (width, height))
    
    counter = 0
    start = time.time() 
    
    while True:
        counter += 1
        ret, frame = capture.read()
        if ret:
            
            seg_frame = self.segmentAsPascalvoc(frame, process_frame=True)
            print("No. of frames:", counter)
            if detect is not None:
              target_class = self.target_obj(detect)
              seg_frame[1][seg_frame[1] != target_class] = 0

            if low == True:
                blur_frame = cv2.blur(frame, (21,21), 0)

            if moderate == True:
                blur_frame = cv2.blur(frame, (39,39), 0)

            if extreme == True:
                blur_frame = cv2.blur(frame, (81,81), 0)

            out = np.where(seg_frame[1], frame, blur_frame)
            

            output = cv2.resize(out, (width,height), interpolation=cv2.INTER_AREA)
            if output_video_name is not None:
                save_video.write(output)

        else:
          break

    capture.release()

    end = time.time()
    print(f"Processed {counter} frames in {end-start:.1f} seconds")
      
    if frames_per_second is not None:
        save_video.release()

    return  output  


   ##### BLUR THE BACKGROUND OF A CAMERA FRAME ######  

  def blur_camera(self, cam, low = False, moderate = False, extreme = False, frames_per_second = None,
   check_fps = False,show_frames = False, frame_name = None, verbose = None, output_video_name = None, detect = None):
    capture = cam
    width = int(capture.get(cv2.CAP_PROP_FRAME_WIDTH))
    height = int(capture.get(cv2.CAP_PROP_FRAME_HEIGHT))
    if frames_per_second is not None:
      save_video = cv2.VideoWriter(output_video_name, cv2.VideoWriter_fourcc(*'DIVX'),frames_per_second, (width, height))
    counter = 0
    start = time.time() 
     
    while True:
        counter += 1
        ret, frame = capture.read()
        if ret:

            seg_frame = self.segmentAsPascalvoc(frame, process_frame=True)
            if detect is not None:
              target_class = self.target_obj(detect)
              seg_frame[1][seg_frame[1] != target_class] = 0
            if verbose is not None:
               print("No.of frames:", counter)
    

            if low == True:
                blur_frame = cv2.blur(frame, (21,21), 0)

            if moderate == True:
                blur_frame = cv2.blur(frame, (39,39), 0)

            if extreme == True:
                blur_frame = cv2.blur(frame, (81,81), 0)

            out = np.where(seg_frame[1], frame, blur_frame)  
            output = cv2.resize(out, (width,height), interpolation=cv2.INTER_AREA)

            if show_frames == True:
                if frame_name is not None:
                    cv2.imshow(frame_name, output)
                    if cv2.waitKey(25) & 0xFF == ord('q'):
                        break 
            if output_video_name is not None:
                save_video.write(output)

        else:

          break
    if check_fps == True:
        out = capture.get(cv2.CAP_PROP_FPS)
        print(f"{out} frames per seconds") 

    capture.release()

    end = time.time()
    if verbose is not None:
      print(f"Processed {counter} frames in {end-start:.1f} seconds")
      
    if frames_per_second is not None:
        save_video.release()

    return  output


   ### GRAYSCALE THE BACKGROUND OF AN IMAGE ###

  def gray_bg(self, image_path, output_image_name = None, verbose = None, detect = None):
    if verbose is not None:
      print("processing image......")
      
    seg_image = self.segmentAsPascalvoc(image_path)

    if detect is not None:
      target_class = self.target_obj(detect)
      seg_image[1][seg_image[1] != target_class] = 0
    
    ori_img = cv2.imread(image_path)
    
    gray_img = cv2.cvtColor(ori_img, cv2.COLOR_BGR2GRAY)
    gray_img = cv2.cvtColor(gray_img, cv2.COLOR_GRAY2BGR)
    result = np.where(seg_image[1], ori_img, gray_img)
    
    if output_image_name is not None:
        cv2.imwrite(output_image_name, result)

    return result



  def gray_frame(self, frame, verbose = None, detect = None):
    if verbose is not None:
      print("processing frame......")
      
    seg_frame = self.segmentAsPascalvoc(frame, process_frame=True)
    if detect is not None:
      target_class = self.target_obj(detect)
      seg_frame[1][seg_frame[1] != target_class] = 0
    
    gray_frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
    gray_frame = cv2.cvtColor(gray_frame, cv2.COLOR_GRAY2BGR)
    result = np.where(seg_frame[1], frame, gray_frame)

    return result  


   ### GRAYSCALE THE BACKGROUND OF A VIDEO ###

  def gray_video(self, video_path, frames_per_second = None, output_video_name = None, detect = None):
    capture = cv2.VideoCapture(video_path)
    width = int(capture.get(cv2.CAP_PROP_FRAME_WIDTH))
    height = int(capture.get(cv2.CAP_PROP_FRAME_HEIGHT))
    if frames_per_second is not None:
      save_video = cv2.VideoWriter(output_video_name, cv2.VideoWriter_fourcc(*'DIVX'),frames_per_second, (width, height))
    
    counter = 0
    start = time.time() 
    
    while True:
        counter += 1
        ret, frame = capture.read()
        if ret:
            seg_frame = self.segmentAsPascalvoc(frame, process_frame=True)
            if detect is not None:
              target_class = self.target_obj(detect)
              seg_frame[1][seg_frame[1] != target_class] = 0

            print("No. of frames:", counter)

            gray_frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
            gray_frame = cv2.cvtColor(gray_frame, cv2.COLOR_GRAY2BGR)
            out = np.where(seg_frame[1], frame, gray_frame)

            output = cv2.resize(out, (width,height), interpolation=cv2.INTER_AREA)
            if output_video_name is not None:
                save_video.write(output)

        else:
          break

    capture.release()

    end = time.time()
    print(f"Processed {counter} frames in {end-start:.1f} seconds")
      
    if frames_per_second is not None:
        save_video.release()

    return  output
      

  ### GRAYSCALE THE BACKGROUND OF A CAMERA FEED ###

  def gray_camera(self, cam, frames_per_second = None, check_fps = False,show_frames = False, 
  frame_name = None, verbose = None, output_video_name = None, detect = None):

    capture = cam
    width = int(capture.get(cv2.CAP_PROP_FRAME_WIDTH))
    height = int(capture.get(cv2.CAP_PROP_FRAME_HEIGHT))
    if frames_per_second is not None:
      save_video = cv2.VideoWriter(output_video_name, cv2.VideoWriter_fourcc(*'DIVX'),frames_per_second, (width, height))
    counter = 0
    start = time.time() 
     
    while True:
        counter += 1
        ret, frame = capture.read()
        if ret:
          seg_frame = self.segmentAsPascalvoc(frame, process_frame=True)
          if detect is not None:
            target_class = self.target_obj(detect)
            seg_frame[1][seg_frame[1] != target_class] = 0

          if verbose is not None:
            print("No. of frames:", counter)

          gray_frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
          gray_frame = cv2.cvtColor(gray_frame, cv2.COLOR_GRAY2BGR)
          out = np.where(seg_frame[1], frame, gray_frame)

          
          output = cv2.resize(out, (width,height), interpolation=cv2.INTER_AREA)
          if show_frames == True:
            if frame_name is not None:
              cv2.imshow(frame_name, output)
              if cv2.waitKey(25) & 0xFF == ord('q'):
                break 
          if output_video_name is not None:
            save_video.write(output)

        else:

          break
    if check_fps == True:
        out = capture.get(cv2.CAP_PROP_FPS)
        print(f"{out} frames per seconds") 

    capture.release()

    end = time.time()
    if verbose is not None:
      print(f"Processed {counter} frames in {end-start:.1f} seconds")
      
    if frames_per_second is not None:
        save_video.release()

    return  output
    


# --------------------------------------------------------
# Fast R-CNN
# Copyright (c) 2015 Microsoft
# Licensed under The MIT License [see LICENSE for details]
# Written by Sergey Karayev
# --------------------------------------------------------

import numpy as np


def compute_overlap(a, b):
    """
    Parameters
    ----------
    a: (N, 4) ndarray of float
    b: (K, 4) ndarray of float
    Returns
    -------
    overlaps: (N, K) ndarray of overlap between boxes and query_boxes
    """
    area = (b[:, 2] - b[:, 0]) * (b[:, 3] - b[:, 1])

    iw = np.minimum(np.expand_dims(a[:, 2], axis=1), b[:, 2]) - np.maximum(np.expand_dims(a[:, 0], 1), b[:, 0])
    ih = np.minimum(np.expand_dims(a[:, 3], axis=1), b[:, 3]) - np.maximum(np.expand_dims(a[:, 1], 1), b[:, 1])

    iw = np.maximum(iw, 0)
    ih = np.maximum(ih, 0)

    ua = np.expand_dims((a[:, 2] - a[:, 0]) * (a[:, 3] - a[:, 1]), axis=1) + area - iw * ih

    ua = np.maximum(ua, np.finfo(float).eps)

    intersection = iw * ih

    return intersection / ua



# -*- coding: utf-8 -*-

"""
keras_resnet.blocks
~~~~~~~~~~~~~~~~~~~

This module implements a number of popular residual blocks.
"""

from ._1d import (
    basic_1d,
    bottleneck_1d
)

from ._2d import (
    basic_2d,
    bottleneck_2d
)

from ._3d import (
    basic_3d,
    bottleneck_3d
)

from ._time_distributed_2d import (
    time_distributed_basic_2d,
    time_distributed_bottleneck_2d
)



"""
MIT License

Copyright (c) 2020 Hyeonki Hong <hhk7734@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import colorsys

# Don't import tensorflow
import cv2
import numpy as np


_HSV = [(1.0 * x / 256, 1.0, 1.0) for x in range(256)]
_COLORS = list(map(lambda x: colorsys.hsv_to_rgb(*x), _HSV))
_COLORS = list(
    map(
        lambda x: (int(x[0] * 255), int(x[1] * 255), int(x[2] * 255)),
        _COLORS,
    )
)
BBOX_COLORS = []
_OFFSET = [0, 8, 4, 12, 2, 6, 10, 14, 1, 3, 5, 7, 9, 11, 13, 15]
for i in range(256):
    BBOX_COLORS.append(_COLORS[(i * 16) % 256 + _OFFSET[(i * 16) // 256]])


def resize_image(
    image: np.ndarray, target_size: int, ground_truth: np.ndarray = None
):
    """
    @param image:        Dim(height, width, channels)
    @param target_size
    @param ground_truth: [[center_x, center_y, w, h, class_id], ...]

    @return resized_image or (resized_image, resized_ground_truth)

    Usage:
        image = media.resize_image(image, yolo.input_size)
        image, ground_truth = media.resize_image(image, yolo.input_size,
                                                                ground_truth)
    """
    height, width, _ = image.shape

    # Resize
    if max(height, width) != target_size:
        scale = min(target_size / width, target_size / height)
        new_width = int(round(width * scale))
        new_height = int(round(height * scale))
        resized_image = cv2.resize(image, (new_width, new_height))
    else:
        new_width = width
        new_height = height
        resized_image = np.copy(image)

    # Pad
    dw = int(target_size - new_width)
    dh = int(target_size - new_height)

    if dw != 0 or dh != 0:
        dw = dw // 2
        dh = dh // 2
        padded_image = np.full(
            (target_size, target_size, 3), 255, dtype=np.uint8
        )
        padded_image[
            dh : new_height + dh, dw : new_width + dw, :
        ] = resized_image
    else:
        padded_image = resized_image

    if ground_truth is None:
        return padded_image

    # Resize ground truth
    ground_truth = np.copy(ground_truth)

    if dw > dh:
        w_h = new_width / new_height
        ground_truth[:, 0] = w_h * (ground_truth[:, 0] - 0.5) + 0.5
        ground_truth[:, 2] = w_h * ground_truth[:, 2]
    elif dw < dh:
        h_w = new_height / new_width
        ground_truth[:, 1] = h_w * (ground_truth[:, 1] - 0.5) + 0.5
        ground_truth[:, 3] = h_w * ground_truth[:, 3]

    return padded_image, ground_truth


def draw_bboxes(image: np.ndarray, bboxes: np.ndarray, classes: dict):
    """
    @parma image:  Dim(height, width, channel)
    @param bboxes: (candidates, 4) or (candidates, 5)
            [[center_x, center_y, w, h, class_id], ...]
            [[center_x, center_y, w, h, class_id, propability], ...]
    @param classes: {0: 'person', 1: 'bicycle', 2: 'car', ...}

    @return drawn_image

    Usage:
        image = media.draw_bboxes(image, bboxes, classes)
    """
    image = np.copy(image)
    height, width, _ = image.shape

    # Set propability
    if bboxes.shape[-1] == 5:
        bboxes = np.concatenate(
            [bboxes, np.full((*bboxes.shape[:-1], 1), 2.0)], axis=-1
        )
    else:
        bboxes = np.copy(bboxes)

    # Convert ratio to length
    bboxes[:, [0, 2]] = bboxes[:, [0, 2]] * width
    bboxes[:, [1, 3]] = bboxes[:, [1, 3]] * height

    # Draw bboxes
    for bbox in bboxes:
        c_x = int(bbox[0])
        c_y = int(bbox[1])
        half_w = int(bbox[2] / 2)
        half_h = int(bbox[3] / 2)
        top_left = (c_x - half_w, c_y - half_h)
        bottom_right = (c_x + half_w, c_y + half_h)
        class_id = int(bbox[4])
        bbox_color = BBOX_COLORS[class_id]
        font_size = 0.4
        font_thickness = 1

        # Draw box
        cv2.rectangle(image, top_left, bottom_right, bbox_color, 2)

        # Draw text box
        bbox_text = "{}: {:.1%}".format(classes[class_id], bbox[5])
        t_size = cv2.getTextSize(bbox_text, 0, font_size, font_thickness)[0]
        cv2.rectangle(
            image,
            top_left,
            (top_left[0] + t_size[0], top_left[1] - t_size[1] - 3),
            bbox_color,
            -1,
        )

        # Draw text
        cv2.putText(
            image,
            bbox_text,
            (top_left[0], top_left[1] - 2),
            cv2.FONT_HERSHEY_SIMPLEX,
            font_size,
            (255 - bbox_color[0], 255 - bbox_color[1], 255 - bbox_color[2]),
            font_thickness,
            lineType=cv2.LINE_AA,
        )

    return image


def read_classes_names(classes_name_path):
    """
    @return {id: class name}
    """
    classes = {}
    with open(classes_name_path, "r") as fd:
        index = 0
        for class_name in fd:
            class_name = class_name.strip()
            if len(class_name) != 0:
                classes[index] = class_name.replace(" ", "_")
                index += 1

    return classes





import tensorflow as tf
import numpy as np
from PIL import Image
from xtreme_vision.Segmentation.deeplab.deeplab import Deeplab_xcep_pascal
from xtreme_vision.Segmentation.deeplab.deeplab import Deeplab_xcep_ade20k
import cv2
import time


class semantic_segmentation():

  def __init__(self):

    self.model = Deeplab_xcep_pascal()

    self.model2 = Deeplab_xcep_ade20k()
      
  def load_pascalvoc_model(self, model_path):
    self.model.load_weights(model_path)
  
  def load_ade20k_model(self, model_path):
    self.model2.load_weights(model_path)   

  def segmentAsPascalvoc(self, image_path, output_image_name=None,overlay=False, verbose = None):            
    trained_image_width=512
    mean_subtraction_value=127.5
    image = np.array(Image.open(image_path))     
   

    # resize to max dimension of images from training dataset
    w, h, n = image.shape

    if n > 3:
      image = cv2.cvtColor(image, cv2.COLOR_BGRA2RGB)
    
    image_overlay = image.copy()

    ratio = float(trained_image_width) / np.max([w, h])
    resized_image = np.array(Image.fromarray(image.astype('uint8')).resize((int(ratio * h), int(ratio * w))))
    resized_image = (resized_image / mean_subtraction_value) -1


    # pad array to square image to match training images
    pad_x = int(trained_image_width - resized_image.shape[0])
    pad_y = int(trained_image_width - resized_image.shape[1])
    resized_image = np.pad(resized_image, ((0, pad_x), (0, pad_y), (0, 0)), mode='constant')

    if verbose is not None:
      print("Processing image....")

    #run prediction
    res = self.model.predict(np.expand_dims(resized_image, 0))
    
    labels = np.argmax(res.squeeze(), -1)
    # remove padding and resize back to original image
    if pad_x > 0:
        labels = labels[:-pad_x]
    if pad_y > 0:
        labels = labels[:, :-pad_y]

    raw_labels = labels
        
    #Apply segmentation color map
    labels = labelP_to_color_image(labels)   
    labels = np.array(Image.fromarray(labels.astype('uint8')).resize((h, w)))
    
    
    new_img = cv2.cvtColor(labels, cv2.COLOR_RGB2BGR)

    if overlay == True:
        alpha = 0.7
        cv2.addWeighted(new_img, alpha, image_overlay, 1 - alpha,0, image_overlay)

        if output_image_name is not None:
          cv2.imwrite(output_image_name, image_overlay)
          print("Processed Image saved successfully in your current working directory.")

        return raw_labels, image_overlay

        
    else:  
        if output_image_name is not None:
  
          cv2.imwrite(output_image_name, new_img)

          print("Processed Image saved successfuly in your current working directory.")

        return raw_labels, new_img 

        
  def segmentAsAde20k(self, image_path, output_image_name=None,overlay=False, verbose = None):            
    trained_image_width=512
    mean_subtraction_value=127.5
    image = np.array(Image.open(image_path))     
    
    # resize to max dimension of images from training dataset
    w, h, n = image.shape

    if n > 3:
      image = cv2.cvtColor(image, cv2.COLOR_BGRA2RGB)
    
    image_overlay = image.copy()

    ratio = float(trained_image_width) / np.max([w, h])
    resized_image = np.array(Image.fromarray(image.astype('uint8')).resize((int(ratio * h), int(ratio * w))))
    resized_image = (resized_image / mean_subtraction_value) -1


    # pad array to square image to match training images
    pad_x = int(trained_image_width - resized_image.shape[0])
    pad_y = int(trained_image_width - resized_image.shape[1])
    resized_image = np.pad(resized_image, ((0, pad_x), (0, pad_y), (0, 0)), mode='constant')

    if verbose is not None:
      print("Processing image....")
    #run prediction
    res = self.model2.predict(np.expand_dims(resized_image, 0))
    
    labels = np.argmax(res.squeeze(), -1)
    # remove padding and resize back to original image
    if pad_x > 0:
        labels = labels[:-pad_x]
    if pad_y > 0:
        labels = labels[:, :-pad_y]

    raw_labels = labels    
        
    #Apply segmentation color map
    labels = labelAde20k_to_color_image(labels)   
    labels = np.array(Image.fromarray(labels.astype('uint8')).resize((h, w)))
    
    
    new_img = cv2.cvtColor(labels, cv2.COLOR_RGB2BGR)

    if overlay == True:
        alpha = 0.7
        cv2.addWeighted(new_img, alpha, image_overlay, 1 - alpha,0, image_overlay)

        if output_image_name is not None:
          cv2.imwrite(output_image_name, image_overlay)
          print("Processed Image saved successfully in your current working directory.")

        return raw_labels, image_overlay 

        
    else:  
        if output_image_name is not None:
  
          cv2.imwrite(output_image_name, new_img)

          print("Processed Image saved successfuly in your current working directory.")

        return raw_labels, new_img 

  def segmentFrameAsPascalvoc(self, frame, output_image_name=None,overlay=False, verbose = None):            
    trained_frame_width=512
    mean_subtraction_value=127.5
       
    frame_overlay = frame.copy()

    # resize to max dimension of images from training dataset
    w, h, _ = frame.shape
    ratio = float(trained_frame_width) / np.max([w, h])
    resized_frame = np.array(Image.fromarray(frame.astype('uint8')).resize((int(ratio * h), int(ratio * w))))
    resized_frame = (resized_frame / mean_subtraction_value) -1


    # pad array to square image to match training images
    pad_x = int(trained_frame_width - resized_frame.shape[0])
    pad_y = int(trained_frame_width - resized_frame.shape[1])
    resized_frame = np.pad(resized_frame, ((0, pad_x), (0, pad_y), (0, 0)), mode='constant')

    if verbose is not None:
      print("Processing frame....")

    #run prediction
    res = self.model.predict(np.expand_dims(resized_frame, 0))
    
    labels = np.argmax(res.squeeze(), -1)
    # remove padding and resize back to original image
    if pad_x > 0:
        labels = labels[:-pad_x]
    if pad_y > 0:
        labels = labels[:, :-pad_y]
        
    raw_labels = labels    
    #Apply segmentation color map
    labels = labelP_to_color_image(labels)   
    labels = np.array(Image.fromarray(labels.astype('uint8')).resize((h, w)))
    
    
    new_frame = cv2.cvtColor(labels, cv2.COLOR_RGB2BGR)

    if overlay == True:
        alpha = 0.7
        cv2.addWeighted(new_frame, alpha, frame_overlay, 1 - alpha,0, frame_overlay)

        if output_image_name is not None:
          cv2.imwrite(output_image_name, frame_overlay)
          print("Processed Image saved successfully in your current working directory.")

        return  raw_labels, frame_overlay

        
    else:  
        if output_image_name is not None:
  
          cv2.imwrite(output_image_name, new_frame)

          print("Processed Image saved successfuly in your current working directory.")

        return raw_labels, new_frame



  def segmentFrameAsAde20k(self, frame, output_image_name=None,overlay=False, verbose = None):            
    trained_frame_width=512
    mean_subtraction_value=127.5     
    frame_overlay = frame.copy()

    # resize to max dimension of images from training dataset
    w, h, _ = frame.shape
    ratio = float(trained_frame_width) / np.max([w, h])
    resized_frame = np.array(Image.fromarray(frame.astype('uint8')).resize((int(ratio * h), int(ratio * w))))
    resized_frame = (resized_frame / mean_subtraction_value) -1


    # pad array to square image to match training images
    pad_x = int(trained_frame_width - resized_frame.shape[0])
    pad_y = int(trained_frame_width - resized_frame.shape[1])
    resized_frame = np.pad(resized_frame, ((0, pad_x), (0, pad_y), (0, 0)), mode='constant')

    if verbose is not None:
      print("Processing frame....")

    #run prediction
    res = self.model2.predict(np.expand_dims(resized_frame, 0))
    
    labels = np.argmax(res.squeeze(), -1)
    # remove padding and resize back to original image
    if pad_x > 0:
        labels = labels[:-pad_x]
    if pad_y > 0:
        labels = labels[:, :-pad_y]

    raw_labels = labels    
    #Apply segmentation color map
    labels = labelAde20k_to_color_image(labels)   
    labels = np.array(Image.fromarray(labels.astype('uint8')).resize((h, w)))
    
    
    new_frame = cv2.cvtColor(labels, cv2.COLOR_RGB2BGR)

    if overlay == True:
        alpha = 0.7
        cv2.addWeighted(new_frame, alpha, frame_overlay, 1 - alpha,0, frame_overlay)

        if output_image_name is not None:
          cv2.imwrite(output_image_name, frame_overlay)
          print("Processed Image saved successfully in your current working directory.")

        return raw_labels, frame_overlay  

        
    else:  
        if output_image_name is not None:
  
          cv2.imwrite(output_image_name, new_frame)

          print("Processed Image saved successfuly in your current working directory.")

        return raw_labels, new_frame      




  
  def process_video_pascalvoc(self, video_path, overlay = False, frames_per_second = None, output_video_name = None):
    capture = cv2.VideoCapture(video_path)
    width = int(capture.get(cv2.CAP_PROP_FRAME_WIDTH))
    height = int(capture.get(cv2.CAP_PROP_FRAME_HEIGHT))
    if frames_per_second is not None:
      save_video = cv2.VideoWriter(output_video_name, cv2.VideoWriter_fourcc(*'DIVX'),frames_per_second, (width, height))
    
    counter = 0
    start = time.time() 


    if overlay == True:
      while True:
        counter += 1
        ret, frame = capture.read()
        if ret:
          frame_copy = frame.copy()
          trained_frame_width = 512
          mean_subtraction_value = 127.5

          # resize to max dimension of images from training dataset
          w, h, _ = frame.shape
          ratio = float(trained_frame_width) / np.max([w, h])
          resized_frame = np.array(Image.fromarray(frame.astype('uint8')).resize((int(ratio * h), int(ratio * w))))
          resized_frame = (resized_frame / mean_subtraction_value) -1


          # pad array to square image to match training images
          pad_x = int(trained_frame_width - resized_frame.shape[0])
          pad_y = int(trained_frame_width - resized_frame.shape[1])
          resized_frame = np.pad(resized_frame, ((0, pad_x), (0, pad_y), (0, 0)), mode='constant')

          pred = self.model.predict(np.expand_dims(resized_frame, axis = 0))
          
          print("No.of frames:", counter)
            
          labels = np.argmax(pred.squeeze(), -1)

          if pad_x > 0:
            labels = labels[:-pad_x]
          if pad_y > 0:
            labels = labels[:, :-pad_y]

          raw_labels = labels  
          segmap = labelP_to_color_image(labels)
          labels = np.array(Image.fromarray(segmap.astype('uint8')).resize((h, w)))
          new_segmap = cv2.cvtColor(labels, cv2.COLOR_BGR2RGB)
            
          alpha = 0.7
          cv2.addWeighted(new_segmap, alpha, frame_copy, 1-alpha,0, frame_copy)
          output = cv2.resize(frame_copy, (width,height), interpolation=cv2.INTER_AREA)
          if output_video_name is not None:
            save_video.write(output)
        else:
          break   

    
      end = time.time()
      print(f"Processed {counter} frames in {end-start:.1f} seconds")
      capture.release()
      if frames_per_second is not None:
        save_video.release()
      return  raw_labels, output

    else:
      while True:
        counter += 1
        ret, frame = capture.read()
        if ret:
          trained_frame_width = 512
          mean_subtraction_value = 127.5

          # resize to max dimension of images from training dataset
          w, h, _ = frame.shape
          ratio = float(trained_frame_width) / np.max([w, h])
          resized_frame = np.array(Image.fromarray(frame.astype('uint8')).resize((int(ratio * h), int(ratio * w))))
          resized_frame = (resized_frame / mean_subtraction_value) -1


          # pad array to square image to match training images
          pad_x = int(trained_frame_width - resized_frame.shape[0])
          pad_y = int(trained_frame_width - resized_frame.shape[1])
          resized_frame = np.pad(resized_frame, ((0, pad_x), (0, pad_y), (0, 0)), mode='constant')

          pred = self.model.predict(np.expand_dims(resized_frame, axis = 0))
          print("No.of frames:", counter)
            
          labels = np.argmax(pred.squeeze(), -1)

          if pad_x > 0:
            labels = labels[:-pad_x]
          if pad_y > 0:
            labels = labels[:, :-pad_y]

          raw_labels = labels  
          segmap = labelP_to_color_image(labels)
          labels = np.array(Image.fromarray(segmap.astype('uint8')).resize((h, w)))
          new_segmap = cv2.cvtColor(labels, cv2.COLOR_BGR2RGB)
            
          output = cv2.resize(new_segmap, (width,height), interpolation=cv2.INTER_AREA)
          if output_video_name is not None:
            save_video.write(output)

        else:
          break

      capture.release()

      end = time.time()
      print(f"Processed {counter} frames in {end-start:.1f} seconds")
      
      if frames_per_second is not None:
        save_video.release()

      return  raw_labels,  output

  

  def process_camera_pascalvoc(self, cam, overlay = False,  check_fps = False, frames_per_second = None, output_video_name = None, show_frames = False, frame_name = None, verbose = None):
    capture = cam
    width = int(capture.get(cv2.CAP_PROP_FRAME_WIDTH))
    height = int(capture.get(cv2.CAP_PROP_FRAME_HEIGHT))
    if frames_per_second is not None:
      save_video = cv2.VideoWriter(output_video_name, cv2.VideoWriter_fourcc(*'DIVX'), frames_per_second, (width, height))
    
    counter = 0
    start = time.time() 
    

    if overlay == True:
      while True:
        counter += 1
        ret, frame = capture.read()
        if ret:
          frame_copy = frame.copy()
          trained_frame_width = 512
          mean_subtraction_value = 127.5

          # resize to max dimension of images from training dataset
          w, h, _ = frame.shape
          ratio = float(trained_frame_width) / np.max([w, h])
          resized_frame = np.array(Image.fromarray(frame.astype('uint8')).resize((int(ratio * h), int(ratio * w))))
          resized_frame = (resized_frame / mean_subtraction_value) -1


          # pad array to square image to match training images
          pad_x = int(trained_frame_width - resized_frame.shape[0])
          pad_y = int(trained_frame_width - resized_frame.shape[1])
          resized_frame = np.pad(resized_frame, ((0, pad_x), (0, pad_y), (0, 0)), mode='constant')

          pred = self.model.predict(np.expand_dims(resized_frame, axis = 0))
          if verbose is not None:
            print("No.of frames:", counter)
            
          labels = np.argmax(pred.squeeze(), -1)

          if pad_x > 0:
            labels = labels[:-pad_x]
          if pad_y > 0:
            labels = labels[:, :-pad_y]

          raw_labels = labels  
          segmap = labelP_to_color_image(labels)
          labels = np.array(Image.fromarray(segmap.astype('uint8')).resize((h, w)))
          new_segmap = cv2.cvtColor(labels, cv2.COLOR_BGR2RGB)
            
          alpha = 0.7
          cv2.addWeighted(new_segmap, alpha, frame_copy, 1-alpha,0, frame_copy)
          output = cv2.resize(frame_copy, (width,height), interpolation=cv2.INTER_AREA)

          if show_frames == True:
            if frame_name is not None:
              cv2.imshow(frame_name, output)
              if cv2.waitKey(25) & 0xFF == ord('q'):
                break 

          if output_video_name is not None:
            save_video.write(output)

         

        else:
          break 

      if check_fps == True:
        out = capture.get(cv2.CAP_PROP_FPS)
        print(f"{out} frames per seconds")

      capture.release()

      end = time.time()
      if verbose is not None:
        print(f"Processed {counter} frames in {end-start:.1f} seconds")
      
      if frames_per_second is not None:
        save_video.release()

      return  raw_labels, output

    else:
      while True:
        counter += 1
        ret, frame = capture.read()
        if ret:
          trained_frame_width = 512
          mean_subtraction_value = 127.5

          # resize to max dimension of images from training dataset
          w, h, _ = frame.shape
          ratio = float(trained_frame_width) / np.max([w, h])
          resized_frame = np.array(Image.fromarray(frame.astype('uint8')).resize((int(ratio * h), int(ratio * w))))
          resized_frame = (resized_frame / mean_subtraction_value) -1


          # pad array to square image to match training images
          pad_x = int(trained_frame_width - resized_frame.shape[0])
          pad_y = int(trained_frame_width - resized_frame.shape[1])
          resized_frame = np.pad(resized_frame, ((0, pad_x), (0, pad_y), (0, 0)), mode='constant')

          pred = self.model.predict(np.expand_dims(resized_frame, axis = 0))
          if verbose is not None:
            print("No.of frames:", counter)
            
          labels = np.argmax(pred.squeeze(), -1)

          if pad_x > 0:
            labels = labels[:-pad_x]
          if pad_y > 0:
            labels = labels[:, :-pad_y]

          raw_labels = labels  
          segmap = labelP_to_color_image(labels)
          labels = np.array(Image.fromarray(segmap.astype('uint8')).resize((h, w)))
          new_segmap = cv2.cvtColor(labels, cv2.COLOR_BGR2RGB)
            

          output = cv2.resize(new_segmap, (width,height), interpolation=cv2.INTER_AREA)

          if show_frames == True:
            if frame_name is not None:
              cv2.imshow(frame_name, output)
              if cv2.waitKey(25) & 0xFF == ord('q'):
                break 

          if output_video_name is not None:
            save_video.write(output)

        else:
          break
      
      if check_fps == True:
        out = capture.get(cv2.CAP_PROP_FPS)
        print("Frame per seconds:", out)

      capture.release()  
        
      end = time.time()
      if verbose is not None:
        print(f"Processed {counter} frames in {end-start:.1f} seconds")
     

      if frames_per_second is not None:
        save_video.release()

      return raw_labels, output 

  


  def process_video_ade20k(self, video_path, overlay = False, frames_per_second = None, output_video_name = None):
    capture = cv2.VideoCapture(video_path)
    width = int(capture.get(cv2.CAP_PROP_FRAME_WIDTH))
    height = int(capture.get(cv2.CAP_PROP_FRAME_HEIGHT))
    if frames_per_second is not None:
      save_video = cv2.VideoWriter(output_video_name, cv2.VideoWriter_fourcc(*'DIVX'), frames_per_second, (width, height))
    
    counter = 0
    start = time.time() 
    

    if overlay == True:
      while True:
        counter += 1
        ret, frame = capture.read()
        if ret:
          frame_copy = frame.copy()
          trained_frame_width = 512
          mean_subtraction_value = 127.5

          # resize to max dimension of images from training dataset
          w, h, _ = frame.shape
          ratio = float(trained_frame_width) / np.max([w, h])
          resized_frame = np.array(Image.fromarray(frame.astype('uint8')).resize((int(ratio * h), int(ratio * w))))
          resized_frame = (resized_frame / mean_subtraction_value) -1


          # pad array to square image to match training images
          pad_x = int(trained_frame_width - resized_frame.shape[0])
          pad_y = int(trained_frame_width - resized_frame.shape[1])
          resized_frame = np.pad(resized_frame, ((0, pad_x), (0, pad_y), (0, 0)), mode='constant')

          pred = self.model2.predict(np.expand_dims(resized_frame, axis = 0))
          print("No.of frames:", counter)
            
          labels = np.argmax(pred.squeeze(), -1)

          if pad_x > 0:
            labels = labels[:-pad_x]
          if pad_y > 0:
            labels = labels[:, :-pad_y]
            
          raw_labels = labels  
          segmap = labelAde20k_to_color_image(labels)
          labels = np.array(Image.fromarray(segmap.astype('uint8')).resize((h, w)))
          new_segmap = cv2.cvtColor(labels, cv2.COLOR_BGR2RGB)
            
          alpha = 0.7
          cv2.addWeighted(new_segmap, alpha, frame_copy, 1-alpha,0, frame_copy)
          output = cv2.resize(frame_copy, (width,height), interpolation=cv2.INTER_AREA)

          if output_video_name is not None:
            save_video.write(output)

        else:
          break  
             
      capture.release()

      end = time.time()
      print(f"Processed {counter} frames in {end-start:.1f} seconds")

      
      if frames_per_second is not None:
        save_video.release()

      return raw_labels, output

    else:
      while True:
        counter += 1
        ret, frame = capture.read()
        if ret:
          frame_copy = frame.copy()
          trained_frame_width = 512
          mean_subtraction_value = 127.5

          # resize to max dimension of images from training dataset
          w, h, _ = frame.shape
          ratio = float(trained_frame_width) / np.max([w, h])
          resized_frame = np.array(Image.fromarray(frame.astype('uint8')).resize((int(ratio * h), int(ratio * w))))
          resized_frame = (resized_frame / mean_subtraction_value) -1


          # pad array to square image to match training images
          pad_x = int(trained_frame_width - resized_frame.shape[0])
          pad_y = int(trained_frame_width - resized_frame.shape[1])
          resized_frame = np.pad(resized_frame, ((0, pad_x), (0, pad_y), (0, 0)), mode='constant')

          pred = self.model2.predict(np.expand_dims(resized_frame, axis = 0))
          print("No.of frames:", counter)
            
          labels = np.argmax(pred.squeeze(), -1)

          if pad_x > 0:
            labels = labels[:-pad_x]
          if pad_y > 0:
            labels = labels[:, :-pad_y]


          raw_labels = labels  
          segmap = labelAde20k_to_color_image(labels)
          labels = np.array(Image.fromarray(segmap.astype('uint8')).resize((h, w)))
          new_segmap = cv2.cvtColor(labels, cv2.COLOR_BGR2RGB)
            
          output = cv2.resize(new_segmap, (width,height), interpolation=cv2.INTER_AREA)

          if output_video_name is not None:
            save_video.write(output)



        else:
          break


      capture.release()  

      end = time.time()
      print(f"Processed {counter} frames in {end-start:.1f} seconds")

      
      if frames_per_second is not None:
        save_video.release()

      return raw_labels,  output    


  




  def process_camera_ade20k(self, cam, overlay = False,  check_fps = False, frames_per_second = None, output_video_name = None, show_frames = False, frame_name = None, verbose = None):
    capture = cam
    width = int(capture.get(cv2.CAP_PROP_FRAME_WIDTH))
    height = int(capture.get(cv2.CAP_PROP_FRAME_HEIGHT))
    if frames_per_second is not None:
      save_video = cv2.VideoWriter(output_video_name, cv2.VideoWriter_fourcc(*'DIVX'), frames_per_second, (width, height))
    
    counter = 0
    start = time.time() 
    

    if overlay == True:
      while True:
        counter += 1
        ret, frame = capture.read()
        if ret:
          frame_copy = frame.copy()
          trained_frame_width = 512
          mean_subtraction_value = 127.5

          # resize to max dimension of images from training dataset
          w, h, _ = frame.shape
          ratio = float(trained_frame_width) / np.max([w, h])
          resized_frame = np.array(Image.fromarray(frame.astype('uint8')).resize((int(ratio * h), int(ratio * w))))
          resized_frame = (resized_frame / mean_subtraction_value) -1


          # pad array to square image to match training images
          pad_x = int(trained_frame_width - resized_frame.shape[0])
          pad_y = int(trained_frame_width - resized_frame.shape[1])
          resized_frame = np.pad(resized_frame, ((0, pad_x), (0, pad_y), (0, 0)), mode='constant')

          pred = self.model2.predict(np.expand_dims(resized_frame, axis = 0))
          if verbose is not None:
            print("No.of frames:", counter)
            
          labels = np.argmax(pred.squeeze(), -1)

          if pad_x > 0:
            labels = labels[:-pad_x]
          if pad_y > 0:
            labels = labels[:, :-pad_y]

          raw_labels = labels  
          segmap = labelAde20k_to_color_image(labels)
          labels = np.array(Image.fromarray(segmap.astype('uint8')).resize((h, w)))
          new_segmap = cv2.cvtColor(labels, cv2.COLOR_BGR2RGB)
            
          alpha = 0.7
          cv2.addWeighted(new_segmap, alpha, frame_copy, 1-alpha,0, frame_copy)
          output = cv2.resize(frame_copy, (width,height), interpolation=cv2.INTER_AREA)

          if show_frames == True:
            if frame_name is not None:
              cv2.imshow(frame_name, output)
              if cv2.waitKey(25) & 0xFF == ord('q'):
                break 

          if output_video_name is not None:
            save_video.write(output)

        else:
          break 

      if check_fps == True:
        out = capture.get(cv2.CAP_PROP_FPS)
        print(f"{out} frames per seconds")    
             
      capture.release()

      end = time.time()
      if verbose is not None:
        print(f"Processed {counter} frames in {end-start:.1f} seconds")

      
      if frames_per_second is not None:
        save_video.release()

      return raw_labels, output

    else:
      while True:
        counter += 1
        ret, frame = capture.read()
        if ret:
          frame_copy = frame.copy()
          trained_frame_width = 512
          mean_subtraction_value = 127.5

          # resize to max dimension of images from training dataset
          w, h, _ = frame.shape
          ratio = float(trained_frame_width) / np.max([w, h])
          resized_frame = np.array(Image.fromarray(frame.astype('uint8')).resize((int(ratio * h), int(ratio * w))))
          resized_frame = (resized_frame / mean_subtraction_value) -1


          # pad array to square image to match training images
          pad_x = int(trained_frame_width - resized_frame.shape[0])
          pad_y = int(trained_frame_width - resized_frame.shape[1])
          resized_frame = np.pad(resized_frame, ((0, pad_x), (0, pad_y), (0, 0)), mode='constant')

          pred = self.model2.predict(np.expand_dims(resized_frame, axis = 0))
          if verbose is not None:
            print("No.of frames:", counter)
            
          labels = np.argmax(pred.squeeze(), -1)

          if pad_x > 0:
            labels = labels[:-pad_x]
          if pad_y > 0:
            labels = labels[:, :-pad_y]

          raw_labels = labels  
          segmap = labelAde20k_to_color_image(labels)
          labels = np.array(Image.fromarray(segmap.astype('uint8')).resize((h, w)))
          new_segmap = cv2.cvtColor(labels, cv2.COLOR_BGR2RGB)
            
          output = cv2.resize(new_segmap, (width,height), interpolation=cv2.INTER_AREA)

          if show_frames == True:
            if frame_name is not None:
              cv2.imshow(frame_name, output)
              if cv2.waitKey(25) & 0xFF == ord('q'):
                break 

          if output_video_name is not None:
            save_video.write(output)



        else:
          break

      if check_fps == True:
        out = capture.get(cv2.CAP_PROP_FPS)
        print("Frame per seconds:", out)   

      capture.release()  

      end = time.time()
      if verbose is not None:
        print(f"Processed {counter} frames in {end-start:.1f} seconds")

      
      if frames_per_second is not None:
          save_video.release()

      return raw_labels,  output    


  
##Create Pascalvoc colormap format ##
    
def create_pascal_label_colormap():
  """Creates a label colormap used in PASCAL VOC segmentation benchmark.

  Returns:
    A Colormap for visualizing segmentation results.
  """
  colormap = np.zeros((256, 3), dtype = int)
  ind = np.arange(256, dtype=int)

  for shift in reversed(range(8)):
    for channel in range(3):
      colormap[:, channel] |= ((ind >> channel) & 1) << shift
    ind >>= 3

  return colormap


#Obtain the results of segmentation#

def obtain_segmentation(image, nc = 21):
    colors = create_pascal_label_colormap()
    r = np.zeros_like(image).astype(np.uint8)
    g = np.zeros_like(image).astype(np.uint8)
    b = np.zeros_like(image).astype(np.uint8)
    
    for a in range(0,nc):
        index = image == a
        r[index] = colors[a, 0]
        g[index] = colors[a, 1]
        b[index] = colors[a, 2]
        rgb = np.stack([r,g,b], axis = 2)  

    return rgb  

# Assign colors to objects #

def labelP_to_color_image(label):
  """Adds color defined by the dataset colormap to the label.

  Args:
    label: A 2D array with integer type, storing the segmentation label.

  Returns:
    result: A 2D array with floating type. The element of the array
      is the color indexed by the corresponding element in the input label
      to the PASCAL color map.

  Raises:
    ValueError: If label is not of rank 2 or its value is larger than color
      map maximum entry.
  """
  if label.ndim != 2:
    raise ValueError('Expect 2-D input label')

  colormap = create_pascal_label_colormap()

  if np.max(label) >= len(colormap):
    raise ValueError('label value too large.')

  return colormap[label]   



##Create Ade20k colormap format ##

def create_ade20k_label_colormap():
  """Creates a label colormap used in ADE20K segmentation benchmark.
  Returns:
    A colormap for visualizing segmentation results.
  """
  return np.asarray([
      [0, 0, 0],
      [120, 120, 120],
      [180, 120, 120],
      [6, 230, 230],
      [80, 50, 50],
      [4, 200, 3],
      [120, 120, 80],
      [140, 140, 140],
      [204, 5, 255],
      [230, 230, 230],
      [4, 250, 7],
      [224, 5, 255],
      [235, 255, 7],
      [150, 5, 61],
      [120, 120, 70],
      [8, 255, 51],
      [255, 6, 82],
      [143, 255, 140],
      [204, 255, 4],
      [255, 51, 7],
      [204, 70, 3],
      [0, 102, 200],
      [61, 230, 250],
      [255, 6, 51],
      [11, 102, 255],
      [255, 7, 71],
      [255, 9, 224],
      [9, 7, 230],
      [220, 220, 220],
      [255, 9, 92],
      [112, 9, 255],
      [8, 255, 214],
      [7, 255, 224],
      [255, 184, 6],
      [10, 255, 71],
      [255, 41, 10],
      [7, 255, 255],
      [224, 255, 8],
      [102, 8, 255],
      [255, 61, 6],
      [255, 194, 7],
      [255, 122, 8],
      [0, 255, 20],
      [255, 8, 41],
      [255, 5, 153],
      [6, 51, 255],
      [235, 12, 255],
      [160, 150, 20],
      [0, 163, 255],
      [140, 140, 140],
      [250, 10, 15],
      [20, 255, 0],
      [31, 255, 0],
      [255, 31, 0],
      [255, 224, 0],
      [153, 255, 0],
      [0, 0, 255],
      [255, 71, 0],
      [0, 235, 255],
      [0, 173, 255],
      [31, 0, 255],
      [11, 200, 200],
      [255, 82, 0],
      [0, 255, 245],
      [0, 61, 255],
      [0, 255, 112],
      [0, 255, 133],
      [255, 0, 0],
      [255, 163, 0],
      [255, 102, 0],
      [194, 255, 0],
      [0, 143, 255],
      [51, 255, 0],
      [0, 82, 255],
      [0, 255, 41],
      [0, 255, 173],
      [10, 0, 255],
      [173, 255, 0],
      [0, 255, 153],
      [255, 92, 0],
      [255, 0, 255],
      [255, 0, 245],
      [255, 0, 102],
      [255, 173, 0],
      [255, 0, 20],
      [255, 184, 184],
      [0, 31, 255],
      [0, 255, 61],
      [0, 71, 255],
      [255, 0, 204],
      [0, 255, 194],
      [0, 255, 82],
      [0, 10, 255],
      [0, 112, 255],
      [51, 0, 255],
      [0, 194, 255],
      [0, 122, 255],
      [0, 255, 163],
      [255, 153, 0],
      [0, 255, 10],
      [255, 112, 0],
      [143, 255, 0],
      [82, 0, 255],
      [163, 255, 0],
      [255, 235, 0],
      [8, 184, 170],
      [133, 0, 255],
      [0, 255, 92],
      [184, 0, 255],
      [255, 0, 31],
      [0, 184, 255],
      [0, 214, 255],
      [255, 0, 112],
      [92, 255, 0],
      [0, 224, 255],
      [112, 224, 255],
      [70, 184, 160],
      [163, 0, 255],
      [153, 0, 255],
      [71, 255, 0],
      [255, 0, 163],
      [255, 204, 0],
      [255, 0, 143],
      [0, 255, 235],
      [133, 255, 0],
      [255, 0, 235],
      [245, 0, 255],
      [255, 0, 122],
      [255, 245, 0],
      [10, 190, 212],
      [214, 255, 0],
      [0, 204, 255],
      [20, 0, 255],
      [255, 255, 0],
      [0, 153, 255],
      [0, 41, 255],
      [0, 255, 204],
      [41, 0, 255],
      [41, 255, 0],
      [173, 0, 255],
      [0, 245, 255],
      [71, 0, 255],
      [122, 0, 255],
      [0, 255, 184],
      [0, 92, 255],
      [184, 255, 0],
      [0, 133, 255],
      [255, 214, 0],
      [25, 194, 194],
      [102, 255, 0],
      [92, 0, 255],
  ])

#Assign colors to objects#  
def labelAde20k_to_color_image(label):
  """Adds color defined by the dataset colormap to the label.

  Args:
    label: A 2D array with integer type, storing the segmentation label.

  Returns:
    result: A 2D array with floating type. The element of the array
      is the color indexed by the corresponding element in the input label
      to the PASCAL color map.

  Raises:
    ValueError: If label is not of rank 2 or its value is larger than color
      map maximum entry.
  """
  if label.ndim != 2:
    raise ValueError('Expect 2-D input label')

  colormap = create_ade20k_label_colormap()

  if np.max(label) >= len(colormap):
    raise ValueError('label value too large.')

  return colormap[label]   

